from PySide6.QtCore import QObject, Signal, Slot, QUrl, QByteArray
from PySide6.QtNetwork import QNetworkAccessManager, QNetworkRequest, QNetworkReply, QSslSocket

from packaging import version

import json
import logging

from ..settings import Settings
from ..app_config import AppConfig
from ..lexemes.lexemes import Lexemes


class UpdateHelper(QObject):
    """
    # How to use example:
    update_helper = UpdateHelper()
    update_helper.new_version_available.connect(lambda v:
                                                QMessageBox.information(None, "Update Available",
                                                                        f"A new version {v} is available."))
    update_helper.check_for_updates()
    """

    STATUS_OK = 'ok'
    STATUS_ERROR = 'error'

    # Signal to emit if a new version is available
    new_version_check_response = Signal(object)  # Or: Signal(str)

    def __init__(self):
        super().__init__()

        self.settings = Settings(parent=self)

        self.logger = logging.getLogger('update_helper')

        self.logging = AppConfig.get_logging()
        self.debug = AppConfig.get_debug()

        if self.logging:
            self.logger.info("SSL supported: %s" % QSslSocket.supportsSsl())

        # Default language setup, change to settings value to modify it via UI
        self.lexemes = Lexemes(self.settings.app_language, default_scope='common')

        self.current_version = version.parse(AppConfig.get_app_version())

        self.manager = QNetworkAccessManager(self)

    def check_for_updates(self):
        # Updates url
        release_url = AppConfig.get_repository_github_release_url()
        request = QNetworkRequest(QUrl(release_url))
        if self.logging:
            self.logger.info(f'Check for update request to {release_url}')
        # Set request data
        # reply = self.manager.post(request, request_data)  # POST
        reply = self.manager.get(request)  # GET

        # Connect finished signal
        reply.finished.connect(lambda _reply=reply: self.handle_network_reply(_reply))
        # Connect error signal
        reply.errorOccurred.connect(lambda error, _reply=reply: self.handle_network_reply(_reply, error))

    @Slot()
    def handle_network_reply(self, reply: QNetworkReply, error_code=None):
        # Get received status code, say 200
        status_code = reply.attribute(QNetworkRequest.Attribute.HttpStatusCodeAttribute)

        if self.debug:
            self.logger.debug(f'Update response, reply: {reply}, error_code: {error_code}, status_code: {status_code}')

        reply.deleteLater()  # Clean up the QNetworkReply object

        # Make sure there is no error
        if reply.error() == QNetworkReply.NetworkError.NoError:
            # The request was successful
            data = reply.readAll()  # type: QByteArray
            """
            # Or like this:
            data = reply.readAll().data().decode('utf-8')
            json_data = json.loads(data)
            """

            if self.debug:
                self.logger.debug(f'Raw RESPONSE [{status_code}]: {data}')

            # json_document = QJsonDocument.fromJson(data)
            # json_data = json_document.object()

            # Convert QByteArray to string
            res_string = data.toStdString()  # Or: str(data.data(), encoding='utf-8')
            if self.debug:
                self.logger.debug(f'Result multi-line STRING: {res_string}')

            # Clean up the string and make one line
            json_str = ''.join(line.strip() for line in res_string.splitlines())

            if self.debug:
                self.logger.debug(f'Result STRING: {json_str}')

            result_message = self.lexemes.get('network_connection_error_empty')
            # Parse JSON response
            try:
                json_data = json.loads(json_str)
                if self.debug:
                    self.logger.debug(f"Result JSON: {json_data}")

                latest_version_str = json_data['tag_name']
                latest_version = version.parse(latest_version_str)

                if latest_version > self.current_version:
                    update_link = '<a href="%s">%s</a>' % (AppConfig.get_repository_github_release_url(), latest_version)
                    result_message = self.lexemes.get('update_helper_new_version_is_available',
                                                      latest_version=update_link)
                    if self.logging:
                        self.logger.info("New version of the app is available: %s. Current version is %s. Response: %s}"
                                         % (latest_version, self.current_version, reply.errorString()))
                    # Emit the new version signal
                    self.new_version_check_response.emit(
                        {'status': self.STATUS_OK, 'msg': result_message,
                         'current_version': self.current_version,
                         # Check new version by this var
                         'new_version': latest_version_str})
                else:
                    result_message = self.lexemes.get('update_helper_latest_version_installed')
                    if self.logging:
                        self.logger.info(
                            "No new version of the app is available, the current version is %s. Response: %s}"
                            % (self.current_version, reply.errorString()))
                    # Emit the same version signal
                    self.new_version_check_response.emit(
                        {'status': self.STATUS_OK, 'msg': result_message, 'current_version': self.current_version})

            except json.JSONDecodeError as e:
                if self.logging:
                    self.logger.warning("Error decoding JSON: %s" % e)
        elif reply.error() == QNetworkReply.NetworkError.HostNotFoundError:
            # The host was not found, indicating possible DNS issues or no internet connection
            result_message = self.lexemes.get('network_connection_error_connection_or_dns')
        elif reply.error() == QNetworkReply.NetworkError.ConnectionRefusedError:
            # Connection was refused, indicating the server might be down or there are network issues
            result_message = self.lexemes.get('network_connection_error_connection_refused')
        elif reply.error() == QNetworkReply.NetworkError.TimeoutError:
            # The connection timed out, indicating network issues
            result_message = self.lexemes.get('network_connection_error_connection_timed_out')
        elif reply.error() == QNetworkReply.NetworkError.ContentNotFoundError:
            # The requested page or resource is not found
            result_message = self.lexemes.get('network_connection_error_connection_404_error')
        else:
            # Handle other errors
            result_message = self.lexemes.get('network_connection_error_generic_with_status_code',
                                              status_code=status_code)

        if error_code is not None:
            if self.logging:
                self.logger.warning(result_message)
                self.logger.warning(f"Failed to fetch update information: {reply.errorString()}")
            # Emit error signal
            self.new_version_check_response.emit({'status': self.STATUS_ERROR, 'msg': result_message})



# Greek lexemes settings_dialog.py
lexemes = {
    # Διάλογος ρυθμίσεων
    "window_title": "Ρυθμίσεις",

    "button_close": "Κλείσιμο",

    "tab_general": "Γενικά",
    "tab_editor_config": "Επεξεργαστής",
    "tab_viewer_config": "Προβολέας",
    "tab_ai_config": "Ρυθμίσεις AI",

    "general_app_config_label": "Ρυθμίσεις Εφαρμογής",
    "general_app_language_label": "Γλώσσα",
    "general_app_language_combo_placeholder_text": "Επιλέξτε γλώσσα",
    "general_app_language_combo_accessible_description": "Γλώσσα διεπαφής της εφαρμογής",
    "general_app_theme_label": "Θέμα",
    "general_app_theme_combo_placeholder_text": "Επιλέξτε θέμα",
    "general_app_theme_combo_accessible_description": "Θέμα διεπαφής της εφαρμογής",
    "general_app_main_menu_label": "Κύριο Μενού",
    "general_app_main_menu_checkbox": "Εμφάνιση Κύριου Μενού",
    "general_app_main_menu_checkbox_accessible_description": "Εμφάνιση του κύριου αναδυόμενου μενού της εφαρμογής",
    "general_app_font_size_label": "Μέγεθος Γραμματοσειράς: {size}pt",
    "general_app_font_size_slider_accessible_description": "Ρύθμιση του παγκόσμιου μεγέθους γραμματοσειράς της εφαρμογής",

    "general_statusbar_label": "Γραμμή Κατάστασης",
    "general_statusbar_show_global_cursor_position_checkbox": "Εμφάνιση Παγκόσμιας Θέσης Δρομέα",
    "general_statusbar_show_global_cursor_position_checkbox_accessible_description":
        "Εμφάνιση της παγκόσμιας θέσης του δρομέα στη γραμμή κατάστασης",

    "editor_config_label": "Ρυθμίσεις Επεξεργαστή",
    "editor_config_show_line_numbers_checkbox": "Εμφάνιση Αριθμών Γραμμών",
    "editor_config_show_line_numbers_checkbox_accessible_description": "Εμφάνιση αριθμών γραμμών στον επεξεργαστή",

    "viewer_config_label": "Ρυθμίσεις Προβολέα",
    "viewer_config_process_emojis_checkbox": "Μετατροπή Εικονιδίων Κειμένου σε Γραφικά",
    "viewer_config_process_emojis_checkbox_accessible_description": "Μετατροπή εικονιδίων κειμένου σε γραφικές αναπαραστάσεις",
    "viewer_config_highlight_todos_checkbox": "Επισήμανση Σημειώσεων TODO",
    "viewer_config_highlight_todos_checkbox_accessible_description": "Επισήμανση ετικετών TODO μέσα στο κείμενο",
    "viewer_config_open_link_confirmation_checkbox": "Απαίτηση Επιβεβαίωσης για Άνοιγμα Συνδέσμων",
    "viewer_config_open_link_confirmation_checkbox_accessible_description":
        "Ζήτηση επιβεβαίωσης πριν το άνοιγμα συνδέσμων",
    "viewer_config_save_resources_checkbox": "Αυτόματη αποθήκευση εξωτερικών εικόνων στον δίσκο",
    "viewer_config_save_resources_checkbox_accessible_description":
        "Αυτόματη αποθήκευση αντιγράφων εξωτερικών εικόνων στον δίσκο για πρόσβαση χωρίς σύνδεση.",

    "ai_config_openai_api_label": "OpenAI API",
    "ai_config_openai_api_url_input_placeholder_text": "URL API",
    "ai_config_openai_api_url_input_accessible_description": "URL του OpenAI API",
    "ai_config_openai_api_key_input_placeholder_text": "Κλειδί API",
    "ai_config_openai_api_key_input_accessible_description": "Κλειδί του OpenAI API",
    "ai_config_openai_api_supported_models_label": "Υποστηριζόμενα μοντέλα",
    "ai_config_ai_model_names_combo_placeholder_text": "Επιλέξτε μοντέλο",
    "ai_config_ai_model_names_combo_accessible_description": "Υποστηριζόμενα μοντέλα προς επιλογή",

    "ai_config_base_label": "Βασικοί Παράμετροι",
    "ai_config_base_system_prompt_label": "Σύστημα Προτροπής",
    "ai_config_base_system_prompt_edit_placeholder_text": "Βασική προτροπή συστήματος που προηγείται κάθε αιτήματος",
    "ai_config_base_system_prompt_edit_accessible_description":
        "Βασική προτροπή συστήματος που προηγείται κάθε αιτήματος. Απλό κείμενο.",
    "ai_config_base_response_max_tokens_label": "Μέγιστοι Διακριτικοί Απόκρισης",
    "ai_config_base_response_max_tokens_input_accessible_description": "Μέγιστος αριθμός διακριτικών για λήψη σε απόκριση",
}


from PySide6.QtCore import Signal
from PySide6.QtGui import QFont
from PySide6.QtWidgets import QTextBrowser

from .app_config import AppConfig

import logging


class ViewWidget(QTextBrowser):
    """
    Class to implement document view functionality.
    It utilises QTextBrowser power by extending or modifying of the original widget methods.
    """

    content_set = Signal()

    def __init__(self, parent=None):
        """
        Args:
            parent (optional): Parent object
        """
        super(ViewWidget, self).__init__(parent)

        # Get app's global font size
        font = QFont()
        font.setPointSize(AppConfig.get_font_size())
        self.setFont(font)

        self.logger = logging.getLogger('view_widget')

        self.logging = AppConfig.get_logging()
        self.debug = AppConfig.get_debug()

    def setDocument(self, document):
        # Override setDocument() to allow additional actions like emit the document set signal
        super().setDocument(document)
        # Emit document set signal to notify listeners
        self.content_set.emit()

    def setHtml(self, html_text):
        # Override setDocument() to allow additional actions like emit the document set signal
        super().setHtml(html_text)
        # Emit document set signal to notify listeners
        self.content_set.emit()

    def mousePressEvent(self, event):
        super().mousePressEvent(event)

        cursor_pos = event.pos()

        anchor = self.anchorAt(cursor_pos)
        # Double check not to process hyperlink click if an expandable block was clicked
        if anchor and anchor.startswith('expandable'):
            if self.debug:
                self.logger.debug("Anchor URL at cursor:", anchor)
            return

    def contextMenuEvent(self, event):
        # Create the standard context menu
        menu = self.createStandardContextMenu()

        # Modify the menu font to hint the size
        menu.setFont(self.font())

        """
        # Optionally add new actions or modify existing ones
        customAction = menu.addAction("Custom Action")
        customAction.triggered.connect(self.custom_action_triggered)
        """

        # Display the menu
        menu.exec_(event.globalPos())

    def loadResource(self, type, url):
        # Log the resource request
        if self.debug:
            self.logger.debug(f"Requesting resource of type {type} at URL: {url.toString()}")



# Notolog
# An open-source markdown editor written in Python.

# Base Syntax Highlighter class tailored for Notolog

"""
This module contains shared functionality for subclasses like the MdHighlighter.
"""

# - GitHub Repository: https://github.com/notolog/notolog-editor
# - PyPI: https://pypi.org/project/notolog
# - WebSite: https://notolog.app
# - Author: Vadim Bakhrenkov
# - Copyright 2024 Vadim Bakhrenkov
# - License: MIT License

from PySide6.QtCore import Qt, QRegularExpression
from PySide6.QtGui import QTextDocument, QSyntaxHighlighter, QTextBlockUserData, QTextBlock, QTextCharFormat
from PySide6.QtGui import QColor, QBrush, QFont

from typing import Union

from . import AppConfig
from . import TextBlockData
from . import ThemeHelper

import logging


class MainHighlighter(QSyntaxHighlighter):
    """
    Syntax highlighter class
    """

    theme = {}
    # Keep it consistent with ini-file name, say 'main.ini', item prefix 'main_color_h1_text'
    theme_ini_prefix = 'main'

    """
    Elements order is matter!
    Say, i* first, b** second, bi*** third one-by-one to override prev token.
        i_, b__, bi___
        code `...`, code ```...```

    nth 0 is a whole string
    nth > 1 calculates as: start = end - length
    """
    re_rules = []

    def __init__(self, document: QTextDocument = None):
        super().__init__(document)

        self.logger = logging.getLogger('highlighter')

        self.logging = AppConfig.get_logging()
        self.debug = False  # AppConfig.get_debug()

        # Get app's global font size
        self.font_size = AppConfig.get_font_size()

        # Walk rules to create QRegExp for each pattern
        self.rules = [(self.get_regex(pattern), nth, tag, group, duple, fmt, reckon)
                      for (pattern, nth, tag, group, duple, fmt, reckon) in self.re_rules]

        # Collect found tokens
        self.tokens = {}
        # Tokens which belong the same line
        self.line_tokens = {}

        # For debug purposes
        self.line_number = None
        self.line_number_log = None

        # Whether a group formatted or not
        self.formatted_group = {}

        # Indicate that the process is a re-highlighting process
        self.rehighlight_block = None

        self.current_block = None  # type: Union[QTextBlock, None]
        self.prev_block = None  # type: Union[QTextBlock, None]

        self.user_data = None  # type: Union[TextBlockData, QTextBlockUserData, None]
        self.prev_user_data = None  # type: Union[TextBlockData, QTextBlockUserData, None]

        self.theme_helper = ThemeHelper()

        self.override_colors()

    def __init_subclass__(cls):
        # Check the required methods are implemented
        if not hasattr(cls, 'highlightBlock'):
            raise NotImplementedError(f'Method highlightBlock() have to be implemented in subclass {cls.__name__}.')

    def override_colors(self):
        for item in self.theme:
            """
            Proof of concept method of how to override the colors set to highlighter's syntax.
            Consider to override the whole map with styles like background pattern.
            """
            # Get color values from the theme helper
            # For example: 'md_color_h1_text', where the 'md' is the theme prefix, and the 'h1_text' is the item
            _color = self.theme_helper.get_color(f'{self.theme_ini_prefix}_color_{item}')
            _background_color = self.theme_helper.get_color(f'{self.theme_ini_prefix}_background_color_{item}')
            if self.debug:
                self.logger.debug(
                    f"Prefix[{self.theme_ini_prefix}] '{item}', color: '{_color}', bg color: '{_background_color}'")
            # At the moment all the colors set in string format, like 'red', 'darkCyan', etc.
            if _color and isinstance(_color, str):
                # Set up color no matter if it was set up or not
                self.theme[item]['color'] = _color
            if _background_color and isinstance(_background_color, str):
                if 'bg' in self.theme[item]:
                    if 'color' in self.theme[item]['bg'] and _background_color:
                        self.theme[item]['bg']['color'] = _background_color
                    elif isinstance(self.theme[item]['bg'], str):
                        self.theme[item]['bg'] = _background_color
                else:
                    # Set up background color no matter if it was set up or not
                    self.theme[item]['bg'] = _background_color

    def get_regex(self, pattern) -> QRegularExpression:
        """
        Get either QRegularExpression or regular Python re
        """
        return QRegularExpression(pattern)

    #
    # Text block style format
    #
    # @staticmethod
    def cf(self, **kwargs):
        """
        Return a QTextCharFormat with style attributes

        Example of color names:
        https://doc.qt.io/qt-6/qcolor.html#predefined-colors
        """

        color = kwargs['color'] if 'color' in kwargs else ''
        style = kwargs['style'] if 'style' in kwargs else ''
        bg = kwargs['bg'] if 'bg' in kwargs else {}
        font_size_ratio = kwargs['font_size_ratio'] if 'font_size_ratio' in kwargs else ''

        if not (color or style or bg or font_size_ratio):
            return

        cformat = QTextCharFormat()

        # Foreground
        if color:
            color_obj = QColor(color)
            cformat.setForeground(color_obj)

        # Background
        if bg:
            bg_color = None  # type: Union[str, None]
            """
            Qt::BrushStyle https://doc.qt.io/qt-6/qt.html#BrushStyle-enum
            """
            bg_pattern = Qt.BrushStyle.SolidPattern
            if type(bg) is str:
                bg_color = bg  # type: ignore
            elif type(bg) is dict:
                if 'color' in bg:
                    bg_color = bg['color']
                if 'pattern' in bg and bg['pattern'] in Qt.BrushStyle:
                    bg_pattern = bg['pattern']
            # Color required
            if bg_color is not None:
                bg_color_obj = QColor(bg_color)
                """
                https://doc.qt.io/qt-6/qbrush.html
                """
                cformat.setBackground(QBrush(bg_color_obj, bg_pattern))

        # Font size
        if font_size_ratio and self.font_size:
            """
            The ratio depends on the font size set in the constructor
            """
            cformat.setFontPointSize(int(self.font_size * font_size_ratio))

        if 'bold' in style:
            cformat.setFontWeight(QFont.Weight.Bold)
        if 'italic' in style:
            cformat.setFontItalic(True)
        if 'underline' in style:
            cformat.setFontUnderline(True)
        if 'strikethrough' in style:
            cformat.setFontStrikeOut(True)
        if 'monospace' in style:
            monospace_font = QFont(
                "font-family: 'Menlo', 'Monaco', 'Consolas', 'Lucida Console', 'Liberation Mono', 'DejaVu Sans Mono', "
                "'Courier New', monospace;")
            monospace_font.setStyleHint(QFont.StyleHint.Monospace)
            # 100% means normal spacing, no extra spacing
            monospace_font.setLetterSpacing(QFont.SpacingType.PercentageSpacing, 100)
            cformat.setFont(monospace_font)

        return cformat

    def rehighlightBlock(self, block):
        if self.debug:
            self.logger.debug('Re-highlighting block data')

        # Leave as default
        self.rehighlight_block = True

        # Initial values
        self.tokens = {}
        self.line_tokens = {}
        self.line_number = None

        super().rehighlightBlock(block)

    def rehighlight(self):
        if self.debug:
            self.logger.debug('Re-highlighting the whole document')

        self.rehighlight_block = False

        # Initial values
        self.tokens = {}
        self.line_tokens = {}
        self.line_number = None

        super().rehighlight()

        self.rehighlight_block = True

    def get_open_close_token_map(self):
        return []

    def get_nl_closing_tokens(self):
        return []

    def highlightBlock(self, text_str):
        """
        Apply a syntax highlighting to each line of the text.
        https://doc.qt.io/qt-6/qsyntaxhighlighter.html#highlightBlock
        """
        pass

    def get_opened_group_token(self, group):
        for token_data in self.get_open_close_token_map():
            if (group == token_data['group']
                    and token_data['open'] in self.tokens
                    and self.tokens[token_data['open']]['o'] is True):
                return token_data['open']
        return None

    def is_any_formatted(self):
        return True if self.formatted_group else False

    def is_group_formatted(self, group):
        return True if group in self.formatted_group else False

    def set_formatted(self, group):
        # token group
        if group in self.formatted_group:
            self.formatted_group[group] += 1
        else:
            self.formatted_group[group] = 1

    def clear_formatted(self):
        self.formatted_group = {}



# This class is intended to help with line numbers area on the left side of the main document editor.
# The bale functions are:
# - Show line/block number for better text recognition.
# - Highlight current line where the cursor located.

# If used with PyQt6 make sure slots and signals are declared correctly, as there are some differences:
# from PyQt6.QtCore import pyqtSlot as Slot

from PySide6.QtCore import Qt, QObject, QRect, QSize, Slot
from PySide6.QtWidgets import QWidget, QTextEdit, QPlainTextEdit
from PySide6.QtGui import QPainter, QColor, QTextFormat, QTextCharFormat, QFont

from typing import Union

from . import Settings
from . import ThemeHelper

from ..edit_widget import EditWidget


class LineNumbers(QWidget):
    def __init__(self, editor: Union[EditWidget, QPlainTextEdit] = None):
        QWidget.__init__(self, editor)

        self.editor = editor

        """
        # Alternatively it can be get from AppConfig:
        font = QFont()
        font.setPointSize(AppConfig.get_font_size())
        self.setFont(font)
        """
        self.setFont(self.editor.document().defaultFont())  # Set global font to the line area

        # Theme helper
        self.theme_helper = ThemeHelper()

        # Settings
        self.settings = Settings(parent=self)

        editor.updateRequest.connect(self.update_request)
        editor.blockCountChanged.connect(self.update_width)
        # TODO set up in settings and find out the better way to highlight the line (when mixing backgrounds)
        # editor.cursorPositionChanged.connect(self.highlight_current_line)

        # Init object with direct updates
        self.update_numbers()

        # self.highlight_current_line()

    def sizeHint(self) -> QSize:
        return QSize(self.line_numbers_width(), 0)

    def paintEvent(self, event) -> None:
        self.line_numbers_paint_event(event)

    @Slot(QRect, int)
    def update_request(self, rect, dy) -> None:
        if dy:
            self.scroll(0, dy)
        else:
            self.update(0, rect.y(), self.width(), rect.height())
        if rect.contains(self.editor.viewport().rect()):
            self.update_width()

    @Slot(int)
    def update_width(self) -> None:
        self.editor.setViewportMargins(self.line_numbers_width(), 0, 0, 0)

    @Slot()
    def highlight_current_line(self) -> None:
        # The text edit widget is not intended to be a read only atm, but check it anyway to make it future-proof
        if not self.editor.isReadOnly():
            selection = QTextEdit.ExtraSelection()  # type: Union[QTextEdit.ExtraSelection, QTextCharFormat, None]
            if not selection:
                return
            selection.cursor = self.editor.textCursor()
            selection.cursor.clearSelection()
            # Set selection properties
            # line_color = QColor(Qt.GlobalColor.green).lighter(175)  # As an example of how to shift base color tone
            line_color = QColor(self.theme_helper.get_color('edit_widget_line_numbers_active_line_background'))
            selection.format.setBackground(line_color)
            selection.format.setProperty(QTextFormat.Property.FullWidthSelection, True)
            # extra_selections.append(selection)
            self.editor.setExtraSelections([selection])

    def update_numbers(self) -> None:
        # Get element parent's contents size as the text edit widget may not be resized yet
        parent_obj = self.editor.parent()  # type: Union[QObject, QWidget]
        element_cr = parent_obj.contentsRect()
        # Update width first
        self.update_width()
        # Or it can be like this: height = self.editor.frameGeometry().height()
        self.setGeometry(
            QRect(element_cr.left(), element_cr.top(), self.line_numbers_width(), element_cr.height())
        )

    def line_numbers_paint_event(self, event) -> None:
        painter = QPainter(self)
        """
        More info about painter https://doc.qt.io/qt-6/qpainter.html#translate-2
        Example: painter.translate(3, 0)

        Global color enum https://doc.qt.io/qt-6/qt.html#GlobalColor-enum
        It can be a color value or pre-set enum's value by name, like: Qt.GlobalColor.transparent
        """
        painter.fillRect(event.rect(), QColor(self.theme_helper.get_color('edit_widget_line_numbers_background')))

        # Get current line number (block) to highlight
        cursor = self.editor.textCursor()
        current_line_block_number = cursor.block().blockNumber()

        block = self.editor.firstVisibleBlock()
        block_number = block.blockNumber()

        top = self.editor.blockBoundingGeometry(block).translated(self.editor.contentOffset()).top()
        bottom = top + self.editor.blockBoundingRect(block).height()

        number_color = QColor(self.theme_helper.get_color('edit_widget_line_numbers_color'))
        # Also: current_number_color = QColor(Qt.GlobalColor.green).lighter(179)
        current_number_color = QColor(self.theme_helper.get_color('edit_widget_line_numbers_active_number'))
        current_number_background = QColor(
            self.theme_helper.get_color('edit_widget_line_numbers_active_number_background'))

        while block.isValid() and top <= event.rect().bottom():
            if block.isVisible() and bottom >= event.rect().top():
                if current_line_block_number == block_number:
                    """
                    Highlight current line number
                    More info of how to fill rectangle area https://doc.qt.io/qt-6/qpainter.html#fillRect-4
                    """
                    painter.fillRect(0, int(top) + 1, self.width(), self.fontMetrics().height(),
                                     current_number_background)
                    painter.setPen(current_number_color)
                else:
                    # Common number color
                    # Also: painter.setPen(Qt.GlobalColor.darkGray)
                    painter.setPen(number_color)

                number = str(block_number + 1)
                # Minus shift to the left to get the right padding from the overall padding
                painter.drawText(-2, int(top) + 1, self.width(),
                                 self.fontMetrics().height(),
                                 Qt.AlignmentFlag.AlignRight, number)
            block = block.next()
            top = bottom
            bottom = top + self.editor.blockBoundingRect(block).height()
            block_number += 1

    def line_numbers_width(self) -> int:
        """
        Get line space width
        @return: int value of space occupied by digits (incl. paddings)
        """
        if not self.visible():
            return 0

        digits = max(2, len(str(self.editor.blockCount())))
        # '5' is an overall value of left and right padding
        space = 5 + self.fontMetrics().horizontalAdvance("0") * digits
        return space

    def visible(self) -> bool:
        return self.settings.show_line_numbers


# tests/test_lexemes.py

from notolog.lexemes.lexemes import Lexemes

from logging import Logger

import pytest
import os


class TestLexemes:

    @pytest.fixture(scope="function", autouse=True)
    def test_obj_lexemes(self, mocker, request):
        """
        Testing object fixture.
        May contain minimal logic for testing purposes.
        """

        # Get the parameter value(s) from the request
        language, lexemes, lexemes_dir = request.param if hasattr(request, 'param') else None

        if lexemes:
            """
            If lexemes explicitly set return them.
            This is for test cases where the lexemes file read is not necessary.
            """
            mocker.patch.object(Lexemes, 'load_lexemes', return_value=lexemes)
        elif lexemes_dir:
            """
            If lexemes dir is set read lexemes from there.
            For test cases to check actual lexemes file read.
            """
            test_lexemes_dir = os.path.join(os.path.dirname(__file__), lexemes_dir)
            mocker.patch.object(Lexemes, 'get_lexemes_dir', return_value=test_lexemes_dir)
        else:
            """
            Return empty lexemes in the rest of the cases.
            """
            mocker.patch.object(Lexemes, 'load_lexemes', return_value={})

        yield Lexemes(language=language)

    @pytest.fixture(scope="function")
    def test_exp_params_fixture(self, request):
        """
        Fixture to pass params and get expected results.
        This method is used only for passing params via pytest fixture.
        """

        # Get the parameter value(s) from the request
        param_values = request.param

        yield param_values

    @pytest.mark.parametrize(
        "test_obj_lexemes, test_exp_params_fixture",
        [
            ((None, None, None), ('en', {})),
            (('', {}, None), ('en', {})),
            (('en', {123: 456}, None), ('en', {123: 456})),
            (('anyval', {'A': 'B', 'C': 'D'}, None), ('anyval', {'A': 'B', 'C': 'D'})),
        ],
        indirect=True
    )
    def test_lexemes_init(self, test_obj_lexemes, test_exp_params_fixture):
        """
        Test and check initial params are exist in the testing object.
        """
        assert isinstance(test_obj_lexemes.logger, Logger)

        exp_language, exp_lexemes = test_exp_params_fixture

        assert hasattr(test_obj_lexemes, 'language')
        assert test_obj_lexemes.language == exp_language
        assert hasattr(test_obj_lexemes, 'lexemes')
        assert test_obj_lexemes.lexemes == exp_lexemes

    @pytest.mark.parametrize(
        "test_obj_lexemes, test_exp_params_fixture",
        [
            ((None, None, None), (None, 'en')),
            (('', None, None), (None, 'en')),
            (('en', None, None), (None, 'en')),
            (('anyval', None, None), (None, 'anyval')),
            (('anyval', None, None), ('new', 'new')),
        ],
        indirect=True
    )
    def test_lexemes_set_language(self, test_obj_lexemes, test_exp_params_fixture):
        """
        Test language setting for lexemes after the one was initialized.
        """
        param_language, exp_language = test_exp_params_fixture

        if param_language:
            test_obj_lexemes.set_language(param_language)

        assert hasattr(test_obj_lexemes, 'language')
        assert test_obj_lexemes.language == exp_language

    @pytest.mark.parametrize(
        "test_obj_lexemes, test_exp_params_fixture",
        [
            ((None, None, None), 'en'),
            (('', None, None), 'en'),
            (('en', None, None), 'en'),
            (('any', None, None), 'any'),
            (('..', None, None), 'en'),
        ],
        indirect=True
    )
    def test_lexemes_get_lexemes_dir(self, mocker, test_obj_lexemes, test_exp_params_fixture):
        """
        Test lexemes target dir path.
        """
        exp_lexemes_dir = test_exp_params_fixture

        mocker.patch.object(os.path, 'dirname', return_value='')

        assert test_obj_lexemes.get_lexemes_dir() == exp_lexemes_dir

    @pytest.mark.parametrize(
        "test_obj_lexemes, test_exp_params_fixture",
        [
            ((None, None, None),
             ('en', {})),
            ((None, None, 'test_lexemes'),
             ('en', {'test_common': {'any': 'thing', 'hello': 'Hello World!', 'test': 'Passed'},
                     'test_scope': {123: 456, 'foo': 'bar', 'test': 'result'}})),
        ],
        indirect=True
    )
    def test_lexemes_load_lexemes(self, mocker, test_obj_lexemes, test_exp_params_fixture):
        """
        Test actual lexemes loaded.
        """
        exp_language, exp_lexemes = test_exp_params_fixture

        assert hasattr(test_obj_lexemes, 'lexemes')
        assert test_obj_lexemes.lexemes == exp_lexemes

    @pytest.mark.parametrize(
        "test_obj_lexemes, test_exp_params_fixture",
        [
            ((None, None, None), (None, None, None)),
            ((None, None, 'test_lexemes'), ('test_common', 'hello', 'Hello World!')),
            ((None, None, 'test_lexemes'), ('test_common', 'test', 'Passed')),
            ((None, None, 'test_lexemes'), ('test_scope', 123, 456)),
            ((None, None, 'test_lexemes'), ('test_scope', 'undefined_name', '')),
            ((None, None, 'test_lexemes'), ('test_scope_undefined', None, None)),
            ((None, None, 'test_lexemes'), ('test_scope_undefined', 'undefined_name', None)),
        ],
        indirect=True
    )
    def test_lexemes_get(self, test_obj_lexemes, test_exp_params_fixture):
        """
        Test getter for correctly loaded and really existent lexemes.
        """
        param_scope, param_name, exp_result = test_exp_params_fixture

        assert test_obj_lexemes.get(param_name, param_scope) == exp_result



from PySide6.QtCore import Signal, QSettings

from .app_config import AppConfig
from .enums.themes import Themes
from .enums.languages import Languages
from .enums.ai_model_names import AiModelNames

from typing import Any, Union

import logging


class Settings(QSettings):
    """
    Class to store app settings params.
    """

    value_changed = Signal(object)  # type: Signal[Union[object, QSettings]]

    def __init__(self, parent=None):
        """
        Args:
            parent (optional): Parent object
        """
        super(Settings, self).__init__(parent)

        self.logger = logging.getLogger('settings')

        self.logging = AppConfig.get_logging()
        self.debug = AppConfig.get_debug()

        self.settings = QSettings()

        if self.debug:
            # Attributes may not be set at the very beginning
            self.logger.debug('Window size %d x %d' % (getattr(self, 'ui_width', 0), getattr(self, 'ui_height', 0)))

        self.value_changed.connect(lambda v: self.settings.sync())

        self.init_fields()

    def init_fields(self):
        # App's UI settings
        self.create_property("app_theme", str, str(Themes.default()))
        self.create_property("app_language", str, str(Languages.default()))
        self.create_property("app_font_size", int, AppConfig.get_font_base_size(),
                             set_prev_func=lambda: AppConfig.set_prev_font_size(self.settings.value("app_font_size")))
        # Window positioning and size
        self.create_property("ui_width", int, 0)
        self.create_property("ui_height", int, 0)
        self.create_property("ui_pos_x", int, 0)
        self.create_property("ui_pos_y", int, 0)
        # Splitter between navigation tree and view/edit widget position
        self.create_property("ui_splitter_pos", int, 0)
        # Global settings
        self.create_property("show_main_menu", bool, True)
        self.create_property("show_deleted_files", bool, False)
        self.create_property("show_global_cursor_position", bool, False)
        # Editor state
        self.create_property("file_path", str, "")
        self.create_property("mode", int, 0)
        self.create_property("source", int, 0)
        self.create_property("line_num", int, 0)
        self.create_property("col_num", int, 0)
        self.create_property("cursor_pos", int, 0)
        self.create_property("viewport_pos", list, [0, 0])  # Scroll doc to this position
        self.create_property("show_line_numbers", bool, True)
        # Viewer state
        self.create_property("viewer_process_emojis", bool, True)
        self.create_property("viewer_highlight_todos", bool, True)
        self.create_property("viewer_open_link_confirmation", bool, True)
        self.create_property("viewer_save_resources", bool, True)
        # Icons to show on the toolbar. Can be checked like:
        # if value is not None and value.isValid() and isinstance(value.toInt()[0], int): ...
        self.create_property("toolbar_icons", int, None)
        self.create_property("enc_iterations", int, None)
        # AI assistant
        self.create_property("ai_config_openai_url", str, "https://api.openai.com/v1/chat/completions")
        self.create_property("ai_config_openai_key", str, "")
        self.create_property("ai_config_openai_model", str, str(AiModelNames.default()))
        self.create_property("ai_config_base_system_prompt", str, "")
        self.create_property("ai_config_base_response_max_tokens", int, 512)

    def create_property(self, param_name: str, param_type: Any, default_value: Any, set_prev_func: Any = None):
        def getter(_self):
            # Type is crucial here
            return param_type(_self.settings.value(param_name, default_value, type=param_type))

        def setter(_self, value):
            if _self.settings.value(param_name) != value:
                if callable(set_prev_func):
                    # Save previous value
                    set_prev_func()
                _self.settings.setValue(param_name, value)
                _self.value_changed.emit({param_name: value})

        # Dynamically create the property and its setter
        setattr(Settings, param_name, property(getter, setter))

    """
    Example of how to set up for each settings property
    """

    @property
    def show_main_menu(self) -> bool:
        # From settings UI dialog
        return self.settings.value("show_main_menu", True, type=bool)  # type: ignore

    @show_main_menu.setter
    def show_main_menu(self, value) -> None:
        self.settings.setValue("show_main_menu", value)
        self.value_changed.emit({'show_main_menu': value})

    def __getattr__(self, name: str) -> Any:
        """
        Handle attribute access dynamically. Helps to avoid undefined attribute error highlighting within IDE.

        When an AttributeError is raised during attribute access, the __getattr__ method is invoked to handle the
        attribute access dynamically. In your case, after the AttributeError is raised and caught by __getattr__,
        the getter() function is invoked as part of the dynamic attribute access.

        Here's how it works:
        1. The code attempts to access an attribute, such as settings.app_font_size.
        2. If the attribute is not found in the object's namespace, Python raises an AttributeError.
        3. The __getattr__ method intercepts the AttributeError and handles it by providing a custom behavior.
           In this case, it raises another AttributeError with a descriptive message.
        4. After the __getattr__ method completes its execution, the original attribute access attempt is retried.
        5. Since the attribute access is retried, Python again tries to access settings.app_font_size.
        6. This time, the attribute settings.app_font_size exists because it was dynamically created using the
           create_property method.
        7. The getter() function associated with the dynamically created property is invoked to retrieve the value of
           app_font_size
        """
        if name in self.__dict__:
            return self.__dict__[name]
        else:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")



# Latin lexemes settings_dialog.py
lexemes = {
    # Configurationes
    "window_title": "Configurationes",

    "button_close": "Claudere",

    "tab_general": "Generale",
    "tab_editor_config": "Editor",
    "tab_viewer_config": "Visor",
    "tab_ai_config": "Configuratio AI",

    "general_app_config_label": "Configuratio App",
    "general_app_language_label": "Lingua",
    "general_app_language_combo_placeholder_text": "Elige linguam",
    "general_app_language_combo_accessible_description": "Lingua interfaciei app",
    "general_app_theme_label": "Thema",
    "general_app_theme_combo_placeholder_text": "Elige thema",
    "general_app_theme_combo_accessible_description": "Thema interfaciei app",
    "general_app_main_menu_label": "Praecipuum Menu",
    "general_app_main_menu_checkbox": "Praecipuum Menu Ostende",
    "general_app_main_menu_checkbox_accessible_description": "Praecipuum dropdown menu applicationis ostendere",
    "general_app_font_size_label": "Magnitudo Fontis: {size}pt",
    "general_app_font_size_slider_accessible_description": "Regula magnitudinem fontis globalis applicationis",

    "general_statusbar_label": "Status Barra",
    "general_statusbar_show_global_cursor_position_checkbox": "Ostende Globalem Cursoris Positionem",
    "general_statusbar_show_global_cursor_position_checkbox_accessible_description":
        "Ostende globalem cursoris positionem in status barra",

    "editor_config_label": "Editoris Configuratio",
    "editor_config_show_line_numbers_checkbox": "Ostende Numeros Linearum",
    "editor_config_show_line_numbers_checkbox_accessible_description": "Ostende numeros linearum in editor",

    "viewer_config_label": "Visoris Configuratio",
    "viewer_config_process_emojis_checkbox": "Converte Textum Emojis ad Graphica",
    "viewer_config_process_emojis_checkbox_accessible_description":
        "Converte emojis textuales in representationes graphicas",
    "viewer_config_highlight_todos_checkbox": "Illustra TODOs",
    "viewer_config_highlight_todos_checkbox_accessible_description": "In textu TODOs signa illustra",
    "viewer_config_open_link_confirmation_checkbox": "Postulat Confirmationem ad Aperiendum Nexus",
    "viewer_config_open_link_confirmation_checkbox_accessible_description":
        "Quaere confirmationem ante apertionem nexus",
    "viewer_config_save_resources_checkbox": "Serva imagines externas in disco automatice",
    "viewer_config_save_resources_checkbox_accessible_description":
        "Automatice servat exemplaria imaginum externarum in disco ad usum sine nexu.",

    "ai_config_openai_api_label": "API OpenAI",
    "ai_config_openai_api_url_input_placeholder_text": "URL API",
    "ai_config_openai_api_url_input_accessible_description": "URL API OpenAI",
    "ai_config_openai_api_key_input_placeholder_text": "Clavis API",
    "ai_config_openai_api_key_input_accessible_description": "Clavis API OpenAI",
    "ai_config_openai_api_supported_models_label": "Modela sustenta",
    "ai_config_ai_model_names_combo_placeholder_text": "Modelum elige",
    "ai_config_ai_model_names_combo_accessible_description": "Modela sustenta ad eligendum",

    "ai_config_base_label": "Parametri Principes",
    "ai_config_base_system_prompt_label": "Systēma Monitum",
    "ai_config_base_system_prompt_edit_placeholder_text": "Monitum systēmatis basis quod præcedit singulas postulationes",
    "ai_config_base_system_prompt_edit_accessible_description":
        "Monitum systēmatis basis quod præcedit singulas postulationes. Textus simpliciter.",
    "ai_config_base_response_max_tokens_label": "Maxima Responsionis Signa",
    "ai_config_base_response_max_tokens_input_accessible_description": "Maximum signorum quæ in responsione accipiuntur",
}


from PySide6.QtCore import Qt, QObject, QRegularExpression
from PySide6.QtWidgets import QDialog, QVBoxLayout, QTabWidget, QWidget, QFrame, QSizePolicy, QPlainTextEdit
from PySide6.QtWidgets import QLabel, QCheckBox, QLineEdit, QPushButton, QComboBox, QSpinBox, QSlider, QSpacerItem

import logging
from typing import Union

from . import AppConfig
from . import Lexemes
from . import ThemeHelper

from ..enums.languages import Languages
from ..enums.ai_model_names import AiModelNames
from ..enums.themes import Themes
from ..ui.enum_combo_box import EnumComboBox


class SettingsDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent, Qt.WindowType.Dialog)

        self.parent = parent

        # Apply font from the dialog instance to the label
        self.setFont(self.parent.font())

        # self.settings = Settings(parent=self)
        # Use parent's settings to allow the Settings emitted signals to be caught
        self.settings = self.parent.settings  # type: ignore

        self.theme_helper = ThemeHelper()

        self.logger = logging.getLogger('settings_dialog')

        self.logging = AppConfig.get_logging()
        self.debug = AppConfig.get_debug()

        # Default language setup, change to settings value to modify it via UI
        self.lexemes = Lexemes(self.settings.app_language, default_scope='settings_dialog')

        self.setWindowTitle(self.lexemes.get('window_title'))

        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        if self.sizeHint().isValid():
            self.setMinimumSize(self.sizeHint())

        # Set dialog size derived from the main window size
        main_window_size = self.parent.size()
        dialog_width = int(main_window_size.width() * 0.33)
        dialog_height = int(main_window_size.height() * 0.33)
        self.setMinimumSize(dialog_width, dialog_height)

        # Tabs widget as a main one
        self.tab_widget = Union[QWidget, None]

        self.setStyleSheet(self.theme_helper.get_css('settings_dialog'))

        self.init_ui()

    def init_ui(self):
        # Dialog widget's main layout
        layout = QVBoxLayout(self)

        # Main tabs widget
        self.tab_widget = QTabWidget()
        self.tab_widget.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        layout.addWidget(self.tab_widget)

        # Add tabs to the main widget
        # General
        tab_general = QWidget(self)
        tab_general.setObjectName('settings_dialog_tab_general')
        # Editor
        tab_editor_config = QWidget(self)
        tab_editor_config.setObjectName('settings_dialog_tab_editor_config')
        # Viewer
        tab_viewer_config = QWidget(self)
        tab_viewer_config.setObjectName('settings_dialog_tab_viewer_config')
        # AI Config
        tab_ai_config = QWidget(self)
        tab_ai_config.setObjectName('settings_dialog_tab_ai_config')

        self.tab_widget.addTab(tab_general, self.lexemes.get('tab_general'))
        self.tab_widget.addTab(tab_editor_config, self.lexemes.get('tab_editor_config'))
        self.tab_widget.addTab(tab_viewer_config, self.lexemes.get('tab_viewer_config'))
        self.tab_widget.addTab(tab_ai_config, self.lexemes.get('tab_ai_config'))

        # Layout for the General tab
        tab_general_layout = QVBoxLayout(tab_general)
        # Layout for the Editor Config tab
        tab_editor_config_layout = QVBoxLayout(tab_editor_config)
        # Layout for the Viewer Config tab
        tab_viewer_config_layout = QVBoxLayout(tab_viewer_config)
        # Layout for the AI Config tab
        tab_ai_config_layout = QVBoxLayout(tab_ai_config)
        """
        Set spacing between widgets if needed:
        tab_ai_config_layout.setSpacing(5)
        """

        fields_conf = [
            # [General]
            # General settings block label
            {"type": QLabel, "name": "settings_dialog_general_app_config_label", "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('general_app_config_label'), "style": {"bold": True},
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Available languages label
            {"type": QLabel, "name": "settings_dialog_general_app_language_label", "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('general_app_language_label'),
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Available languages dropdown list
            {"type": EnumComboBox, "args": [sorted(Languages, key=lambda member: (not member.is_default, member.value))],
             "name": "settings_dialog_general_app_language_combo:app_language",  # Lexeme key : Object name
             "callback": lambda obj: tab_general_layout.addWidget(obj),
             "placeholder_text": self.lexemes.get('general_app_language_combo_placeholder_text'),
             "accessible_description":
                 self.lexemes.get('app_language_combo_accessible_description')},
            # Available themes label
            {"type": QLabel, "name": "settings_dialog_general_app_theme_label", "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('general_app_theme_label'),
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Available themes dropdown list
            {"type": EnumComboBox, "args": [sorted(Themes, key=lambda member: (not member.is_default, member.value))],
             "name": "settings_dialog_general_app_theme_combo:app_theme",  # Lexeme key : Object name
             "callback": lambda obj: tab_general_layout.addWidget(obj),
             "placeholder_text": self.lexemes.get('general_app_theme_combo_placeholder_text'),
             "accessible_description":
                 self.lexemes.get('general_app_theme_combo_accessible_description')},
            # Horizontal spacer
            {"type": self.layout_horizontal_spacer, "args": [tab_general_layout]},
            # Main menu label
            {"type": QLabel, "name": "settings_dialog_general_app_main_menu_label",
             "alignment": Qt.AlignmentFlag.AlignLeft, "style": {"bold": True},
             "text": self.lexemes.get('general_app_main_menu_label'),
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Either show or not main menu
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_general_app_main_menu_checkbox:show_main_menu",
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('general_app_main_menu_checkbox'),
             "accessible_description":
                 self.lexemes.get('general_app_main_menu_checkbox_accessible_description')},
            # Horizontal spacer
            {"type": self.layout_horizontal_spacer, "args": [tab_general_layout]},
            # Status bar settings block label
            {"type": QLabel, "name": "settings_dialog_general_statusbar_label", "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('general_statusbar_label'), "style": {"bold": True},
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Either show or not global position at status bar
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_general_statusbar_show_global_cursor_position_checkbox"
                     ":show_global_cursor_position",
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('general_statusbar_show_global_cursor_position_checkbox'),
             "accessible_description":
                 self.lexemes.get('general_statusbar_show_global_cursor_position_checkbox_accessible_description')},
            # Horizontal spacer
            {"type": self.layout_horizontal_spacer, "args": [tab_general_layout]},
            # Main menu label
            {"type": QLabel, "name": "settings_dialog_general_app_font_size_label",
             "alignment": Qt.AlignmentFlag.AlignLeft, "style": {"bold": True},
             "text": self.lexemes.get('general_app_font_size_label', size=self.settings.app_font_size),
             "callback": lambda obj: tab_general_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            {"type": QSlider, "args": [Qt.Orientation.Horizontal],
             "props": {'setFocusPolicy': Qt.FocusPolicy.StrongFocus, 'setTickPosition': QSlider.TickPosition.TicksAbove,
                       'setTickInterval': 5, 'setSingleStep': 1,
                       'setMinimum': AppConfig.get_font_min_size(), 'setMaximum': AppConfig.get_font_max_size()},
             "name": "settings_dialog_general_app_font_size_slider:app_font_size",  # Lexeme key : Object name
             "callback": lambda obj: tab_general_layout.addWidget(obj),
             "accessible_description":
                 self.lexemes.get('general_app_font_size_slider_accessible_description')},
            # Spacer to keep elements above on top
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding),
             "callback": lambda obj: tab_general_layout.addWidget(obj)},

            # [Editor config]
            # Editor block label
            {"type": QLabel, "name": "settings_dialog_editor_config_label", "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('editor_config_label'), "style": {"bold": True},
             "callback": lambda obj: tab_editor_config_layout.addWidget(obj)},
            # Either to show or not editor's line numbers
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_editor_config_show_line_numbers_checkbox:show_line_numbers",
             "callback": lambda obj: tab_editor_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('editor_config_show_line_numbers_checkbox'),
             "accessible_description":
                 self.lexemes.get('editor_config_show_line_numbers_checkbox_accessible_description')},
            # Spacer to keep elements above on top
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding),
             "callback": lambda obj: tab_editor_config_layout.addWidget(obj)},

            # [Viewer config]
            # Viewer block label
            {"type": QLabel, "name": "settings_dialog_viewer_config_label", "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('viewer_config_label'), "style": {"bold": True},
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj)},
            # Either to show or not viewer's emojis
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_viewer_config_process_emojis_checkbox:viewer_process_emojis",
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('viewer_config_process_emojis_checkbox'),
             "accessible_description":
                 self.lexemes.get('viewer_config_process_emojis_checkbox_accessible_description')},
            # Spacer
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum),
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj)},
            # Either to show or not viewer's TODOs
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_viewer_config_highlight_todos_checkbox:viewer_highlight_todos",
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('viewer_config_highlight_todos_checkbox'),
             "accessible_description":
                 self.lexemes.get('viewer_config_highlight_todos_checkbox_accessible_description')},
            # Spacer
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum),
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj)},
            # Either to show open link confirmation dialog or not
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_viewer_config_open_link_confirmation_checkbox:viewer_open_link_confirmation",
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('viewer_config_open_link_confirmation_checkbox'),
             "accessible_description":
                 self.lexemes.get('viewer_config_open_link_confirmation_checkbox_accessible_description')},
            # Spacer
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum),
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj)},
            # Auto save downloaded resources on disk
            {"type": QCheckBox,
             # Lexeme key : Object name
             "name": "settings_dialog_viewer_config_save_resources_checkbox:viewer_save_resources",
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "text": self.lexemes.get('viewer_config_save_resources_checkbox'),
             "accessible_description":
                 self.lexemes.get('viewer_config_save_resources_checkbox_accessible_description')},
            # Spacer to keep elements above on top
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding),
             "callback": lambda obj: tab_viewer_config_layout.addWidget(obj)},

            # [AI config]
            # OpenAI API block label
            {"type": QLabel, "name": "settings_dialog_ai_config_openai_api_label", "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('ai_config_openai_api_label'), "style": {"bold": True},
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj)},
            # OpenAI API url line input
            {"type": QLineEdit, "name": "settings_dialog_ai_config_openai_url:ai_config_openai_url",
             "read_only": False, "max_length": 128,
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "placeholder_text": self.lexemes.get('ai_config_openai_api_url_input_placeholder_text'),
             "accessible_description":
                 self.lexemes.get('ai_config_openai_api_url_input_accessible_description')},
            # OpenAI API key line input
            {"type": QLineEdit, "name": "settings_dialog_ai_config_openai_key:ai_config_openai_key",
             "read_only": False, "max_length": 128,
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "placeholder_text": self.lexemes.get('ai_config_openai_api_key_input_placeholder_text'),
             "accessible_description":
                 self.lexemes.get('ai_config_openai_api_key_input_accessible_description')},
            # Spacer
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum),
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj)},
            # Supported models label
            {"type": QLabel, "name": "settings_dialog_ai_config_openai_api_supported_models_label",
             "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('ai_config_openai_api_supported_models_label'),
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj)},
            # Supported models dropdown list
            {"type": EnumComboBox, "args": [sorted(AiModelNames, key=lambda member: member.legacy)],  # legacy below
             "name": "settings_dialog_ai_config_ai_model_names_combo:ai_config_openai_model",
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj),
             "placeholder_text": self.lexemes.get('ai_config_ai_model_names_combo_placeholder_text'),
             "accessible_description":
                 self.lexemes.get('ai_config_ai_model_names_combo_accessible_description')},
            # Horizontal spacer
            {"type": self.layout_horizontal_spacer, "args": [tab_ai_config_layout]},
            # Status bar settings block label
            {"type": QLabel, "name": "settings_dialog_ai_config_base_label", "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('ai_config_base_label'), "style": {"bold": True},
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Base system prompt label
            {"type": QLabel, "name": "settings_dialog_ai_config_base_system_prompt_label",
             "alignment": Qt.AlignmentFlag.AlignLeft, "text": self.lexemes.get('ai_config_base_system_prompt_label'),
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Base system prompt text edit
            {"type": QPlainTextEdit, "name": "settings_dialog_ai_config_base_system_prompt:ai_config_base_system_prompt",
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "placeholder_text": self.lexemes.get('ai_config_base_system_prompt_edit_placeholder_text'),
             "accessible_description":
                 self.lexemes.get('ai_config_base_system_prompt_edit_accessible_description'), "text_lines": 7},
            # Spacer
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum),
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj)},
            # Base response max tokens label
            {"type": QLabel, "name": "settings_dialog_ai_config_base_response_max_tokens_label",
             "alignment": Qt.AlignmentFlag.AlignLeft,
             "text": self.lexemes.get('ai_config_base_response_max_tokens_label'),
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop)},
            # Base response max tokens input
            {"type": QSpinBox, "props": {'setMinimum': 1, 'setMaximum': 65536},  # Update the highest range
             "size_policy": (QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred),
             "name": "settings_dialog_ai_config_base_response_max_tokens:ai_config_base_response_max_tokens",
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj, alignment=Qt.AlignmentFlag.AlignTop),
             "accessible_description":
                 self.lexemes.get('ai_config_base_response_max_tokens_input_accessible_description')},
            # Spacer to keep elements above on top
            {"type": QWidget, "name": None, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding),
             "callback": lambda obj: tab_ai_config_layout.addWidget(obj)},
        ]

        for conf in fields_conf:
            self.create_setting_field(conf)

        # Close button
        close_button = QPushButton(self.lexemes.get('button_close'))
        close_button.setObjectName('settings_dialog_button_close')
        close_button.clicked.connect(self.close)
        layout.addWidget(close_button)

        self.connect_widgets()

        # Adjust size policy to allow for dynamic resizing
        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        self.adjustSize()

        # Get the preferred size of the dialog's content
        preferred_size = self.layout().sizeHint()
        # Adjust the dialog's size based on the preferred size then
        if preferred_size.isValid():
            self.resize(preferred_size)

    def apply_props(self, widget, properties):
        for prop, value in properties.items():
            # Get the method based on the property name
            setter = getattr(widget, prop, None)
            if callable(setter):
                setter(value)  # Call the setter with the provided value

    def create_setting_field(self, conf) -> Union[QCheckBox, QLineEdit, QPlainTextEdit, QComboBox, QSpinBox, QSlider]:
        """
        Method to create setting field with ease.
        @param conf: Config of the field. As the param bounded to the particular method it will be called then.
        @return: Any object instance
        """
        args = conf['args'] if 'args' in conf else [self]
        obj = conf['type'](*args)  # type: conf['type']
        # Align with the dialog font size (mostly for QLabel)
        if hasattr(obj, 'setFont'):
            obj.setFont(self.font())
        if hasattr(obj, 'sizeHint'):
            obj.sizeHint()
        # Apply methods from props
        if 'props' in conf:
            self.apply_props(obj, conf['props'])
        if 'name' in conf and conf['name']:
            """
            Let's set up object name to make it easier to understand any further callbacks later.
            [ Keep object's name consistent with the corresponding variable in Settings class to link them. ]
            and the var name for convenient reading and in case of refactoring.
            """
            obj.setObjectName(conf['name'])
        if 'text' in conf:
            obj.setText(conf['text'])
        if 'read_only' in conf:
            obj.setReadOnly(conf['read_only'])
        if 'max_length' in conf:
            obj.setMaxLength(conf['max_length'])
        if 'placeholder_text' in conf:
            obj.setPlaceholderText(conf['placeholder_text'])
        if 'accessible_description' in conf:
            obj.setAccessibleDescription(conf['accessible_description'])
        if 'size_policy' in conf:
            params = conf['size_policy']
            if isinstance(params, tuple):
                obj.setSizePolicy(*params)  # Contains a few params
            else:
                obj.setSizePolicy(params)
        if 'alignment' in conf:
            obj.setAlignment(conf['alignment'])
        if 'callback' in conf and callable(conf['callback']):
            conf['callback'](obj)
        if 'style' in conf:
            styles = []
            if 'bold' in conf['style']:
                styles.append('font-weight: bold;')  # Bold
            if 'italic' in conf['style']:
                styles.append('font-style: italic;')  # Italic
            if 'color' in conf['style']:
                styles.append(f"color: {conf['style']['color']};")  # Color
            if styles:
                obj.setStyleSheet('QObject { %s }' % ' '.join(styles))
        if 'text_lines' in conf and isinstance(obj, QPlainTextEdit) and int(conf['text_lines']) > 0:
            font_metrics = obj.fontMetrics()
            # Calculate the height needed for N lines based on font metrics
            lines_height = font_metrics.height() * conf['text_lines']
            obj.setFixedHeight(lines_height)
        # Return created object
        return obj

    def parse_object_name(self, object_name: str):
        if object_name.__contains__(":"):
            object_name_parts = object_name.split(":")
            if len(object_name_parts) == 2:
                # lexeme_key, setting_name
                return object_name_parts
            else:
                if self.logging:
                    self.logger.warning(f"Object name in a wrong format {object_name}")
        else:
            # lexeme_key, setting_name
            return object_name, object_name

    def connect_widgets(self):
        # Find children of type QCheckBox
        checkboxes = self.findChildren(QCheckBox)
        for checkbox in checkboxes:
            if isinstance(checkbox, QCheckBox):
                # Parse object name in case it consists of composed data of lexeme and setting names
                _lexeme_key, setting_name = self.parse_object_name(checkbox.objectName())
                if hasattr(self.settings, setting_name):
                    checkbox.setChecked(getattr(self.settings, setting_name, False))
                # Connect signal after set up defaults or restore saved value to avoid signal emitting right after.
                checkbox.stateChanged.connect(self.save_settings)

        # Find children of type QComboBox
        combo_boxes = self.findChildren(QComboBox)
        for combo_box in combo_boxes:
            if isinstance(combo_box, EnumComboBox):
                # Parse object name in case it consists of composed data of lexeme and setting names
                _lexeme_key, setting_name = self.parse_object_name(combo_box.objectName())
                if hasattr(self.settings, setting_name):
                    setting_value = getattr(self.settings, setting_name, None)
                    # Get the index of the Enum member by the name stored in the settings (e.g. Language's 'EN')
                    index = next((i for i, val in enumerate(combo_box.enum_class) if val.name.lower() == setting_value),
                                 None)
                    if self.debug:
                        self.logger.debug(f'Enum {combo_box.enum_class} index for value: {index} [{setting_value}]')
                    if index is not None:
                        combo_box.setCurrentIndex(index)
                # Connect signal after set up defaults or restore saved value to avoid signal emitting right after.
                combo_box.currentIndexChanged.connect(self.save_settings)

        # Find children of type QSpinBox
        spin_boxes = self.findChildren(QSpinBox)
        for spin_box in spin_boxes:
            if isinstance(spin_box, QSpinBox):
                # Parse object name in case it consists of composed data of lexeme and setting names
                _lexeme_key, setting_name = self.parse_object_name(spin_box.objectName())
                if hasattr(self.settings, setting_name):
                    spin_box.setValue(getattr(self.settings, setting_name, 0))
                spin_box.valueChanged.connect(self.save_settings)

        # Find children of type QLineEdit
        line_edits = self.findChildren(QLineEdit)
        for line_edit in line_edits:
            if isinstance(line_edit, QLineEdit):
                # Parse object name in case it consists of composed data of lexeme and setting names
                _lexeme_key, setting_name = self.parse_object_name(line_edit.objectName())
                if hasattr(self.settings, setting_name):
                    line_edit.setText(str(getattr(self.settings, setting_name, '')))
                # Connect signal after set up defaults or restore saved value to avoid signal emitting right after.
                line_edit.textChanged.connect(self.save_settings)

        # Find children of type QPlainTextEdit
        text_edits = self.findChildren(QPlainTextEdit)
        for text_edit in text_edits:
            if isinstance(text_edit, QPlainTextEdit):
                # Parse object name in case it consists of composed data of lexeme and setting names
                _lexeme_key, setting_name = self.parse_object_name(text_edit.objectName())
                if hasattr(self.settings, setting_name):
                    text_edit.setPlainText(getattr(self.settings, setting_name, ''))
                # Connect signal after set up defaults or restore saved value to avoid signal emitting right after.
                text_edit.textChanged.connect(self.save_settings)

        # Find children of type QSpinBox
        sliders = self.findChildren(QSlider)
        for slider in sliders:
            if isinstance(slider, QSlider):
                # Parse object name in case it consists of composed data of lexeme and setting names
                _lexeme_key, setting_name = self.parse_object_name(slider.objectName())
                if hasattr(self.settings, setting_name):
                    slider.setValue(getattr(self.settings, setting_name, 0))
                slider.valueChanged.connect(self.save_settings)

    def save_settings(self):
        # Signal sender
        sender = self.sender()

        # Determine which widget emitted the signal by the object name set
        sender_widget = self.sender()
        sender_name = sender_widget.objectName()

        # Parse object name in case it consists of composed data of lexeme and setting names
        lexeme_key, setting_name = self.parse_object_name(sender_name)

        setting_value = None
        setting_text = None
        if isinstance(sender_widget, QCheckBox):
            setting_value = sender_widget.isChecked()
            setting_text = sender_widget.text()
        elif isinstance(sender_widget, QComboBox):
            # Index of the item in the enum
            # index = sender_widget.currentIndex()
            # setting_value = sender_widget.currentData()
            setting_text = sender_widget.currentText()
            """
            To get enum name instead of the value, say for Languages:
            index = sender_widget.currentIndex()
            setting_value = list(Languages)[index].name
            """
            setting_value = sender_widget.currentData().name.lower()  # Save in lower case
        elif isinstance(sender_widget, QSpinBox):
            setting_value = sender_widget.value()
            setting_text = sender_widget.text()
        elif isinstance(sender_widget, QLineEdit):
            setting_value = sender_widget.text()
            setting_text = sender_widget.placeholderText()
        elif isinstance(sender_widget, QPlainTextEdit):
            setting_value = sender_widget.toPlainText()
            setting_text = sender_widget.placeholderText()
        elif isinstance(sender_widget, QSlider):
            setting_value = sender_widget.value()
            setting_text = sender_widget.accessibleDescription()

        if self.debug:
            self.logger.debug(f"Saving setting '{setting_name}': {setting_value} ({setting_text})")

        try:
            setattr(self.settings, setting_name, setting_value)
        except AttributeError as e:
            if self.logging:
                self.logger.warning(f'ERROR: {e}')

        if self.debug:
            self.logger.debug(f"Setting new value: {getattr(self.settings, setting_name)}")

        """
        In case of language change try to update as much text labels as possible.
        It may not cover all the labels, but may help with update anyway.
        """
        if setting_name == 'app_language' or sender.objectName().endswith('app_language'):
            # Update lexemes object as app language has just been changed
            self.lexemes = Lexemes(self.settings.app_language, default_scope='settings_dialog')
            # Get all the new lexemes
            app_language_lexemes = self.lexemes.get_all()
            # Update dialog title at first
            self.setWindowTitle(self.lexemes.get('window_title'))
            # Iterate each lexeme's scope to match it with object name to update the text if applicable
            for scope in app_language_lexemes:
                if self.debug:
                    self.logger.debug(f'Lexeme scope to check updates {scope}')
                # Iterate each lexeme from the scope
                for lexeme_key in app_language_lexemes[scope].keys():
                    """
                    Some objects may have name set with ":" delimiter, which means "lexeme_key:setting_name".
                    It may help to transfer more params with object name with ease.
                    """
                    # Parse object name in case it consists of composed data of lexeme and setting names
                    lexeme_key, setting_name = self.parse_object_name(lexeme_key)

                    # Get lexeme in the new language set
                    lexeme = app_language_lexemes[scope][lexeme_key]

                    # Search with regex as the object name should contain scope and may contain setting name as well,
                    # but the lexeme contains a key only.
                    regex = QRegularExpression(f"(?={scope}_)?{lexeme_key}.*?")
                    # self.findChildren() contains only the elements related to this dialog
                    found_objects = self.findChildren(QObject, regex, Qt.FindChildOption.FindChildrenRecursively)
                    if not found_objects:
                        # Try to find it in the parent dialog
                        found_objects = self.parent.findChildren(QObject, regex,
                                                                 Qt.FindChildOption.FindChildrenRecursively)
                        if self.debug and found_objects:
                            self.logger.debug(f"Found object to update ({scope}){lexeme_key}: {found_objects}")

                    # Search by full object name
                    # object_name = self.lexemes.get_full_key(lexeme_key)
                    # found_objects = self.findChildren(QObject, object_name, Qt.FindChildrenRecursively)

                    for obj in found_objects:
                        if ((isinstance(obj, (QLabel, QCheckBox, QComboBox, QPushButton)))
                                and hasattr(obj, 'setText')
                                and callable(getattr(obj, 'setText'))):
                            if self.debug:
                                self.logger.debug(
                                    f'Object lexeme update: {obj.objectName()} with id: {id(obj)}, lexeme "{lexeme}"')
                            obj.setText(lexeme)
                        if isinstance(obj, QWidget):
                            # Add iterable map of the tabs to change the search by QWidget to explicitly set tab text.
                            self.set_tab_text(obj.objectName(), lexeme)
                        # Exclusions
                        if obj.objectName() == "settings_dialog_general_app_font_size_label":
                            obj.setText(
                                self.lexemes.get('general_app_font_size_label', size=self.settings.app_font_size))

        if setting_name == 'app_theme' or sender.objectName().endswith('app_theme'):
            self.theme_helper = ThemeHelper()
            # Update self styles
            self.setStyleSheet(self.theme_helper.get_css('settings_dialog'))

        if setting_name == 'app_font_size' or sender.objectName().endswith('app_font_size'):
            # Update font from parent as it should be updated and incorporates new font size as well.
            # Simple update of the font size should also work.
            self.setFont(self.parent.font())
            widgets_to_update = [QLabel, QTabWidget, QPushButton, QCheckBox, QLineEdit, QPlainTextEdit, QComboBox,
                                 QSpinBox, QSlider]
            for _widget in widgets_to_update:
                # Find all QLabel objects
                found_objects = self.findChildren(_widget)
                for obj in found_objects:
                    # Align with the dialog font size (mostly for QLabel)
                    if hasattr(obj, 'setFont'):
                        # obj.sizeHint()
                        obj.setFont(self.font())
                    if obj.objectName() == "settings_dialog_general_app_font_size_label":
                        """
                        # Update the font size on the label
                        font = QFont()
                        font.setPointSize(setting_value)
                        app_font_size_label.setFont(font)
                        """
                        obj.setText(self.lexemes.get('general_app_font_size_label', size=setting_value))

    def set_tab_text(self, object_name, text):
        # Iterate through all the tabs
        for index in range(self.tab_widget.count()):
            # Get the widget for the current index
            widget = self.tab_widget.widget(index)

            # Check if this widget matches the object name
            if widget.objectName() == object_name:
                # Set the text
                self.tab_widget.setTabText(index, text)
                break

    def layout_horizontal_spacer(self, layout: QVBoxLayout):
        # Add spacer above the line
        layout.addSpacerItem(QSpacerItem(0, 10, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed))

        # Add horizontal delimiter
        line = QFrame()
        line.setFrameShape(QFrame.Shape.HLine)
        line.setFrameShadow(QFrame.Shadow.Sunken)
        layout.addWidget(line)

        # Add spacer below the line
        layout.addSpacerItem(QSpacerItem(0, 10, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed))



# French lexemes settings_dialog.py
lexemes = {
    # Paramètres
    "window_title": "Paramètres",

    "button_close": "Fermer",

    "tab_general": "Général",
    "tab_editor_config": "Éditeur",
    "tab_viewer_config": "Visualiseur",
    "tab_ai_config": "Config IA",

    "general_app_config_label": "Configuration de l'application",
    "general_app_language_label": "Langue",
    "general_app_language_combo_placeholder_text": "Choisir une langue",
    "general_app_language_combo_accessible_description": "Langue de l'interface de l'application",
    "general_app_theme_label": "Thème",
    "general_app_theme_combo_placeholder_text": "Choisir un thème",
    "general_app_theme_combo_accessible_description": "Thème de l'interface de l'application",
    "general_app_main_menu_label": "Menu Principal",
    "general_app_main_menu_checkbox": "Afficher le menu principal",
    "general_app_main_menu_checkbox_accessible_description": "Afficher le menu déroulant principal de l'application",
    "general_app_font_size_label": "Taille de police : {size}pt",
    "general_app_font_size_slider_accessible_description": "Ajuster la taille de la police globale de l'application",

    "general_statusbar_label": "Barre de statut",
    "general_statusbar_show_global_cursor_position_checkbox": "Afficher la position globale du curseur",
    "general_statusbar_show_global_cursor_position_checkbox_accessible_description":
        "Afficher la position globale du curseur dans la barre de statut",

    "editor_config_label": "Configuration de l'éditeur",
    "editor_config_show_line_numbers_checkbox": "Afficher les numéros de ligne",
    "editor_config_show_line_numbers_checkbox_accessible_description": "Afficher les numéros de ligne dans l'éditeur",

    "viewer_config_label": "Configuration du visualiseur",
    "viewer_config_process_emojis_checkbox": "Convertir les emojis de texte en graphiques",
    "viewer_config_process_emojis_checkbox_accessible_description":
        "Convertir les emojis de texte en représentations graphiques",
    "viewer_config_highlight_todos_checkbox": "Mettre en évidence les TODO",
    "viewer_config_highlight_todos_checkbox_accessible_description": "Souligner les balises TODO dans le texte",
    "viewer_config_open_link_confirmation_checkbox": "Demander confirmation avant d'ouvrir les liens",
    "viewer_config_open_link_confirmation_checkbox_accessible_description":
        "Demander une confirmation avant d'ouvrir des liens",
    "viewer_config_save_resources_checkbox": "Enregistrer automatiquement les images externes sur le disque",
    "viewer_config_save_resources_checkbox_accessible_description":
        "Enregistre automatiquement des copies des images externes sur le disque pour un accès hors ligne.",

    "ai_config_openai_api_label": "API OpenAI",
    "ai_config_openai_api_url_input_placeholder_text": "URL de l'API",
    "ai_config_openai_api_url_input_accessible_description": "URL de l'API OpenAI",
    "ai_config_openai_api_key_input_placeholder_text": "Clé API",
    "ai_config_openai_api_key_input_accessible_description": "Clé de l'API OpenAI",
    "ai_config_openai_api_supported_models_label": "Modèles pris en charge",
    "ai_config_ai_model_names_combo_placeholder_text": "Choisir un modèle",
    "ai_config_ai_model_names_combo_accessible_description": "Modèles pris en charge à choisir",

    "ai_config_base_label": "Paramètres de base",
    "ai_config_base_system_prompt_label": "Invite de système",
    "ai_config_base_system_prompt_edit_placeholder_text": "Invite de système de base qui précède chaque demande",
    "ai_config_base_system_prompt_edit_accessible_description":
        "Invite de système de base qui précède chaque demande. Texte brut.",
    "ai_config_base_response_max_tokens_label": "Nombre maximal de jetons de réponse",
    "ai_config_base_response_max_tokens_input_accessible_description": "Nombre maximal de jetons à recevoir en réponse",
}


# Russian lexemes ai_assistant.py
lexemes = {
    # ИИ помощник
    "dialog_title": "Запрос к ИИ",
    "dialog_prompt_input_placeholder_text": "Запрос",
    "dialog_prompt_input_accessible_description": "Введите запрос, чтобы дать контекст ИИ-помощнику",
    "dialog_response_output_placeholder_text": "Данные ответа",
    "dialog_response_output_accessible_description": "Данные ответа появятся в этом поле",
    "dialog_response_output_notice_empty_text": "Введите текст, чтобы получить ответ примерно...",

    "dialog_usage_model_label": "МОДЕЛЬ",
    "dialog_usage_tokens_label": "ТОКЕНЫ",
    "dialog_usage_tokens_prompt": "запрос: {tokens}",
    "dialog_usage_tokens_answer": "ответ: {tokens}",
    "dialog_usage_tokens_total": "всего: {tokens}",
    "dialog_button_send_request": "Отправить запрос",
}


# English lexemes ai_assistant.py
lexemes = {
    # AI assistant dialog
    "dialog_title": "Request to AI",
    "dialog_prompt_input_placeholder_text": "Request",
    "dialog_prompt_input_accessible_description": "Input request to give a context to the AI-assistant",
    "dialog_response_output_placeholder_text": "Response data",
    "dialog_response_output_accessible_description": "Response data will appear at this field",
    "dialog_response_output_notice_empty_text": "Provide a text input to get a response of about...",

    "dialog_usage_model_label": "MODEL",
    "dialog_usage_tokens_label": "TOKENS",
    "dialog_usage_tokens_prompt": "prompt: {tokens}",
    "dialog_usage_tokens_answer": "answer: {tokens}",
    "dialog_usage_tokens_total": "total: {tokens}",
    "dialog_button_send_request": "Send Request",
}


from PySide6.QtCore import Qt
from PySide6.QtWidgets import QDialog, QLabel, QVBoxLayout, QDialogButtonBox, QSizePolicy
from PySide6.QtGui import QFontMetrics

from typing import Optional, Callable, Any

from . import Settings
from . import AppConfig
from . import Lexemes

import logging


class EncPasswordResetDialog(QDialog):

    def __init__(self, callback: Optional[Callable[..., Any]] = None, parent=None):
        super().__init__(parent)

        self.parent = parent

        if self.parent and hasattr(self.parent, 'font'):
            # Apply font from the main window to the dialog
            self.setFont(self.parent.font())

        self.settings = Settings(parent=self)

        self.logger = logging.getLogger('enc_password_reset_dialog')

        self.logging = AppConfig.get_logging()
        self.debug = AppConfig.get_debug()

        # Default language setup, change to settings value to modify it via UI
        self.lexemes = Lexemes(self.settings.app_language, default_scope='common')

        title = self.lexemes.get('dialog_encrypt_password_reset_title')
        self.setWindowTitle(title)
        self.setObjectName('dialog_encrypt_password_reset_title')

        # Adjust dialog size
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        if self.sizeHint().isValid():
            self.setMinimumSize(self.sizeHint())

        label = QLabel()
        label.setObjectName('dialog_encrypt_password_reset_text')
        label.setText(self.lexemes.get('dialog_encrypt_password_reset_text'))
        label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        if label.sizeHint().isValid():
            label.setMinimumSize(label.sizeHint())
        label.adjustSize()

        button_box = QDialogButtonBox()
        button_box.setObjectName('dialog_encrypt_password_reset_button_box')
        button_box.setOrientation(Qt.Orientation.Horizontal)
        # Check translations
        button_box.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Yes)
        button_box.setCenterButtons(True)

        button_cancel = button_box.button(QDialogButtonBox.StandardButton.Cancel)
        button_cancel.setText(self.lexemes.get('dialog_encrypt_password_reset_button_cancel'))
        button_yes = button_box.button(QDialogButtonBox.StandardButton.Yes)
        button_yes.setText(self.lexemes.get('dialog_encrypt_password_reset_button_yes'))

        # not `self` to avoid: 'Attempting to add QLayout "" to ... "", which already has a layout
        layout = QVBoxLayout()
        layout.addWidget(label)
        layout.addWidget(button_box)
        self.setLayout(layout)

        self.adjust_minimum_width(title)

        button_box.accepted.connect(lambda: callback(callback=self.reject) if callable(callback) else self.reject)
        button_box.rejected.connect(self.reject)

    def adjust_minimum_width(self, title: str) -> None:
        # Calculate width required for the title text
        font_metrics = QFontMetrics(self.font())
        title_width = font_metrics.horizontalAdvance(title)

        # Set minimum width for the dialog
        self.setMinimumWidth(title_width + 150)  # Add some padding



# Notolog
# An open-source markdown editor written in Python.

# Main file to start up the app and to set up an async loop.

"""
This module is the main entry point for the Notolog app. It handles the initialization of the
application environment, processes command line arguments, and starts the main application loop.

Detailed Description:
- Initializes the GUI by calling the main module.
- Set up global logging and debug settings.
- Initializes the async loop.

Usage:
Run the module directly from the command line with the necessary arguments:
python3 main.py
"""

# - GitHub Repository: https://github.com/notolog/notolog-editor
# - PyPI: https://pypi.org/project/notolog
# - WebSite: https://notolog.app
# - Author: Vadim Bakhrenkov
# - Copyright 2024 Vadim Bakhrenkov
# - License: MIT License


from PySide6.QtWidgets import QStyleFactory
from qasync import QEventLoop, QApplication

# from notolog.app_config import AppConfig
from notolog.notolog_editor import NotologEditor

import platform
import logging
import asyncio
import sys
import os

# Force Qt style override
os.environ["QT_STYLE_OVERRIDE"] = "Fusion"


def main():
    """
    Possible params:
    filename='notolog.log'
    format='%(asctime)s:%(levelname)s:%(name)s:%(message)s'
    """
    logging.basicConfig(level=logging.DEBUG, format='[%(name)s] %(funcName)s: %(levelname)s: %(message)s')

    # Enable logging, uncomment or change app config
    # AppConfig.set_logging(True)
    # Enable debug mode (a lot of logs), uncomment or change app config
    # AppConfig.set_debug(False)

    # Main application
    app = QApplication(sys.argv)

    # Detect the operating system to choose the style
    current_os = platform.system()
    if current_os == "Windows":
        app.setStyle(QStyleFactory.create("WindowsVista"))
    # elif current_os == "Darwin":  # MacOS
    #    app.setStyle(QStyleFactory.create("Macintosh"))  # Renders not as expected
    else:  # Or: current_os == "Linux"
        app.setStyle(QStyleFactory.create("Fusion"))  # Fusion is a cross-platform choice

    # Maintain a unique style regardless of the user's system settings
    app.setDesktopSettingsAware(False)

    logging.getLogger('notolog').info('Application dir path "%s"' % QApplication.applicationDirPath())

    # Get the screen to pass it to the main module
    screen = app.screens()[0]

    # Main loop
    loop = QEventLoop(app)
    asyncio.set_event_loop(loop)

    app_close_event = asyncio.Event()
    app.aboutToQuit.connect(app_close_event.set)

    # Start up the editor
    editor = NotologEditor(screen=screen)
    editor.show()

    with loop:
        """
        Contains run_forever() which is contain app.exec(),
        No need to run sys.exit(app.exec()) in this case.
        """
        loop.run_until_complete(app_close_event.wait())


if __name__ == '__main__':
    main()



from PySide6.QtCore import Qt, QModelIndex
from PySide6.QtGui import QColor
from PySide6.QtWidgets import QFileSystemModel

import os

from typing import Any

from . import ThemeHelper

from ..helpers.file_helper import remove_trailing_numbers


class FileSystemModel(QFileSystemModel):
    def __init__(self, *args, **kwargs):
        super(FileSystemModel, self).__init__(*args, **kwargs)

        self.condition = None
        self.color = None

        # Theme helper
        self.theme_helper = ThemeHelper()

        self.highlighted_indexes = []

    def highlight(self, index, condition, color):
        """
        Keep in mind: the method has invoked on prev items atm.
        @param index: File index
        @param condition: File condition (name in this case)
        @param color: QColor for the role
        @return: None
        """
        self.condition = condition
        self.color = color

        if index.isValid():
            if not any(d.get('condition') == self.condition for d in self.highlighted_indexes):
                self.highlighted_indexes.append({'index': index, 'condition': condition})
                if len(self.highlighted_indexes) > 3:
                    # Or: self.highlighted_indexes = self.highlighted_indexes[-3:]
                    self.highlighted_indexes.pop(0)  # Store only the last X of them, remove the others
            # Data changed signal
            self.dataChanged.emit(index, index, [Qt.ItemDataRole.BackgroundRole])

    def clear_highlights(self):
        """
        Clear last X elements in the file tree. It may help when previous elements were highlighted special way.
        @return: None
        """
        for idx in self.highlighted_indexes:
            index = idx['index']
            condition = idx['condition']
            # Params the data() will be called with
            self.condition = condition
            # Allow to re-draw the role with default fallback color
            self.color = self.theme_helper.get_color('main_tree_background', True)
            # Emit the change
            self.dataChanged.emit(index, index, [Qt.ItemDataRole.BackgroundRole])
        self.highlighted_indexes.clear()

    def data(self, index=int, role=Qt.ItemDataRole.DisplayRole) -> Any:
        """
        Items view decoration.
        More info about ItemDataRole https://doc.qt.io/qt-6/qt.html#ItemDataRole-enum
        """

        # Item's background decoration
        if self.condition and role == Qt.ItemDataRole.BackgroundRole:
            text = index.data(Qt.ItemDataRole.DisplayRole)
            """
            Search 'like' match:
            if self.condition in text:
            """
            # Search exact match
            if self.condition == text:
                return QColor(self.color)

        # Item's text color decoration
        if role == Qt.ItemDataRole.ForegroundRole:
            file_path = self.filePath(index)
            if remove_trailing_numbers(file_path).endswith('del'):
                color = self.theme_helper.get_color('main_tree_file_type_del')
                return QColor(color)

        # Item's icon decoration
        if role == Qt.ItemDataRole.DecorationRole:
            info = self.fileInfo(index)
            """
            Decorate file's tree icon by file extension
            """
            # TODO more variants
            if os.path.isfile(info.filePath()):
                """
                Match case statement works for Python 3.10 upwards
                match info.suffix():
                    case 'md':
                        return ...
                Most common switch case
                """
                if info.suffix() == 'md':
                    color = self.theme_helper.get_color('main_tree_file_type_md')
                    return self.theme_helper.get_icon(theme_icon='filetype-md.svg', color=QColor(color))
                elif info.suffix() == 'txt':
                    color = self.theme_helper.get_color('main_tree_file_type_txt')
                    return self.theme_helper.get_icon(theme_icon='filetype-txt.svg', color=QColor(color))
                elif info.suffix() == 'html':
                    color = self.theme_helper.get_color('main_tree_file_type_html')
                    return self.theme_helper.get_icon(theme_icon='filetype-html.svg', color=QColor(color))
                elif info.suffix() == 'enc':
                    color = self.theme_helper.get_color('main_tree_file_type_enc')
                    return self.theme_helper.get_icon(theme_icon='file-earmark-lock2.svg', color=QColor(color))
                elif remove_trailing_numbers(info.suffix()) == 'del':
                    color = self.theme_helper.get_color('main_tree_file_type_del')
                    return self.theme_helper.get_icon(theme_icon='file-earmark-x.svg', color=QColor(color))
                else:
                    color = self.theme_helper.get_color('main_tree_file_type_default')
                    return self.theme_helper.get_icon(theme_icon='file-earmark.svg', color=QColor(color))
            elif os.path.isdir(info.filePath()):
                if info.fileName() == "..":
                    color = self.theme_helper.get_color('main_tree_folder_dotdot')
                    return self.theme_helper.get_icon(theme_icon='folder-symlink.svg', color=QColor(color))
                else:
                    color = self.theme_helper.get_color('main_tree_folder')
                    return self.theme_helper.get_icon(theme_icon='folder.svg', color=QColor(color))

        return super(FileSystemModel, self).data(index, role)



# tests/test_notolog_editor.py

from notolog.notolog_editor import NotologEditor
from notolog.settings import Settings
from notolog.editor_state import Mode
from notolog.lexemes.lexemes import Lexemes

import pytest

import os


class TestNotologEditor:

    @pytest.fixture(scope="function")
    def test_obj_lexemes(self, mocker):
        # Return empty lexemes
        mocker.patch.object(Lexemes, 'load_lexemes', return_value={})
        # Init lexemes
        _lexemes = Lexemes()
        # Mock get lexeme results
        mocker.patch.object(_lexemes, 'get', return_value='Lorem ipsum')

        yield _lexemes

    @pytest.fixture(scope="function")
    def test_obj_notolog_editor(self, mocker, test_obj_lexemes):
        # Mock __init__ method
        mocker.patch.object(NotologEditor, '__init__', return_value=None)

        _obj = NotologEditor()
        # Set lexemes object
        setattr(_obj, 'lexemes', test_obj_lexemes)

        yield _obj

    @pytest.fixture(scope="function")
    def test_obj_settings(self):
        yield Settings()

    def test_notolog_editor_load_default_page(self, mocker, test_obj_notolog_editor, test_obj_settings, tmp_path):
        """
        Check that the default page is loaded
        @param mocker: pytest fixture
        @param test_obj_notolog_editor: NotologEditor instance
        @param test_obj_settings: Settings instance
        @param tmp_path: fixture to get the path to a temporary directory
        @return: None
        """
        test_obj_notolog_editor.settings = test_obj_settings
        assert isinstance(test_obj_notolog_editor.settings, Settings)

        # Specify a side effect method for the mocked object
        mocker.patch('os.path.isfile', side_effect=lambda filename: filename.endswith('.md'))

        mock_method1 = mocker.patch.object(test_obj_notolog_editor, 'get_any_file', return_value=None)
        mock_method2 = mocker.patch.object(test_obj_notolog_editor, 'load_file', return_value=True)

        # Call the method under test
        result = test_obj_notolog_editor.load_default_page()
        assert result is True

        # Current working directory
        # test_run_from_dir = os.getcwd()
        test_file_dir = os.path.dirname(os.path.realpath(__file__))
        test_file_parent_dir = os.path.dirname(test_file_dir)
        # Assert that the method was called once
        mock_method1.assert_not_called()
        # Assert that the method was called once with the param(s)
        mock_method2.assert_called_once_with('%s/README.md' % test_file_parent_dir)

    def test_notolog_editor_load_default_page_any(self, mocker, test_obj_notolog_editor, test_obj_settings):
        """
        Check that an any page is loaded instead of default page
        @param mocker: pytest fixture
        @param test_obj_notolog_editor: NotologEditor instance
        @param test_obj_settings: Settings instance
        @return: None
        """
        # test_obj_notolog_editor.settings = mocker.Mock()
        test_obj_notolog_editor.settings = test_obj_settings
        assert isinstance(test_obj_notolog_editor.settings, Settings)

        mocker.patch.object(os.path, 'isfile', return_value=False)

        mock_method1 = mocker.patch.object(test_obj_notolog_editor, 'get_any_file', return_value='Test_File_Path.Py')
        mock_method2 = mocker.patch.object(test_obj_notolog_editor, 'load_file', return_value=True)

        # Call the method under test
        result = test_obj_notolog_editor.load_default_page()
        assert result is True

        # Assert that the method was called once
        mock_method1.assert_called_once()
        # Assert that the method was called once with the param(s)
        mock_method2.assert_called_once_with('Test_File_Path.Py')

    @pytest.fixture(scope="function")
    def test_exp_params_fixture(self, request):
        # Get the parameter value(s) from the request
        param_values = request.param

        yield param_values

    @pytest.mark.parametrize(
        "test_exp_params_fixture",
        [
            (Mode.VIEW, 'new-document-1.md', None, None, 1),
            (Mode.VIEW, 'new-document-1.md', False, None, 1),
            (Mode.VIEW, 'new-document-9999.md', True, None, 9999),
            (Mode.VIEW, 'new-document-1.md', False, 'Lorem ipsum', 1),
        ],
        indirect=True
    )
    def test_notolog_editor_action_new_file(self, mocker, test_obj_notolog_editor, test_exp_params_fixture):
        mode, file_path, isfile, content, res_path_call_cnt = test_exp_params_fixture

        mocker.patch.object(test_obj_notolog_editor, 'get_mode', return_value=mode)
        mocker.patch.object(test_obj_notolog_editor, 'toggle_mode', return_value=None)
        setattr(test_obj_notolog_editor, 'debug', False)

        mock_res_path = mocker.patch.object(test_obj_notolog_editor, 'get_active_dir', return_value=file_path)
        #mock_res_path = mocker.patch('app.notolog_editor.res_path', return_value=file_path)
        # fixture arg: monkeypatch
        # monkeypatch.setattr('app.notolog_editor.res_path', lambda _file_path: _file_path == file_path)
        mocker.patch.object(os.path, 'isfile', return_value=isfile)

        mock_save_file_content = mocker.patch.object(test_obj_notolog_editor, 'save_file_content', return_value=None)
        mocker.patch.object(test_obj_notolog_editor, 'load_file', return_value=None)

        test_obj_notolog_editor.action_new_file(content)

        assert mock_res_path.call_count == res_path_call_cnt
        # Check the parameters passed to the mocked function(s)
        # assert str(mock_res_path.call_args) == "call('%s')" % file_path
        # assert str(mock_save_file_content.call_args) == "call('%s', '%s')" % (file_path, content)
        if content:
            assert str(content) in str(mock_save_file_content.call_args)



from enum import Enum


class EnumBase(Enum):

    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        """
        Override _generate_next_value_ to allow for custom attributes on enum members (mostly for default).

        'name' is the name of the Enum's member, which will be 'A', 'B', 'C', etc.
        This method returns the name itself and a 'default' marker (False)
        """
        return name, False

    def __str__(self):
        return self.name.lower()

    @classmethod
    def __getitem__(cls, key):
        if isinstance(key, str):
            key = key.upper()  # Normalize the key to uppercase
        return cls._member_map_[key]

    @classmethod
    def default(cls):
        """
        Returns the default value of the Enum.
        Usage:
            self.default()
        @return: Any
        """
        # Find and return the default enum member
        for member in cls:
            if hasattr(member, 'is_default') and member.is_default:
                return member
        return None  # If no default is marked, return None or consider raising an error



from PySide6.QtCore import Qt
from PySide6.QtWidgets import QApplication
from PySide6.QtTest import QTest

from notolog.notolog_editor import NotologEditor
from notolog.enums.languages import Languages
from notolog.settings import Settings

import os
import sys
import pytest


class TestQtUi:

    @pytest.fixture(autouse=True, scope="session")
    def qt_app(self):
        # Check to avoid:
        # RuntimeError: Please destroy the QApplication singleton before creating a new QApplication instance.
        app = QApplication.instance()
        if not app:
            # Force Qt style override
            """
            QApplication: invalid style override 'kvantum' passed, ignoring it.
                Available styles: Windows, Fusion
            """
            os.environ["QT_STYLE_OVERRIDE"] = "Fusion"
            # Fixture to initialize QApplication.
            app = QApplication(sys.argv)

        yield app

        # Properly close the QApplication after each test
        app.quit()

    @pytest.fixture
    def settings_obj(self, qt_app):
        # Fixture to create and return settings instance
        settings = Settings()
        yield settings

    @pytest.fixture
    def main_window(self, mocker, qt_app):
        # Force to override system language as a default
        mocker.patch.object(Languages, 'default', return_value='la')

        # Fixture to create and return main window instance
        window = NotologEditor(screen=qt_app.screens()[0])
        yield window

    def test_editor_state(self, main_window, settings_obj):
        # Check app language set correctly
        assert settings_obj.app_language == 'la'

        # Check default window title
        assert main_window.windowTitle() == "Editorium Notolog"

        assert main_window.statusbar['mode_label'].text() == 'Modus Visum'
        assert main_window.statusbar['source_label'].text() == 'HTML'
        assert main_window.statusbar['encryption_label'].text() == 'Simplicitas 🔓'

        # Test that clicking the edit button updates the editor state
        QTest.mouseClick(main_window.toolbar.toolbar_edit_button, Qt.MouseButton.LeftButton)

        assert main_window.statusbar['mode_label'].text() == 'Modus Editio'
        assert main_window.statusbar['source_label'].text() == 'Markdown'
        assert main_window.statusbar['encryption_label'].text() == 'Simplicitas 🔓'

    def test_search_elements(self, main_window):
        assert hasattr(main_window.toolbar, 'search_input')
        main_window.toolbar.search_input.setText('Test search')
        assert main_window.toolbar.search_input.text() == 'Test search'

        assert hasattr(main_window.toolbar, 'btn_search_clear')
        # Test that clicking the button updates the search field
        QTest.mouseClick(main_window.toolbar.btn_search_clear, Qt.MouseButton.LeftButton)
        assert main_window.toolbar.search_input.text() == ''



from PySide6.QtCore import Qt, QSize
from PySide6.QtGui import QAction, QColor
from PySide6.QtWidgets import QToolBar, QWidget, QMenu, QLabel, QLineEdit, QCheckBox, QToolButton, QPushButton
from PySide6.QtWidgets import QSizePolicy

from . import Settings
from . import AppConfig
from . import Lexemes
from . import ThemeHelper

from typing import Union

import logging


class ToolBar(QToolBar):
    # Main window toolbar class

    def __init__(self, parent=None, actions=None, buttons=None, refresh=None):
        """
        Args:
            parent (optional): Parent object
            actions (List[Dict[str, Any]], optional): The map with the toolbar's icons
            buttons (List[Dict[str, Any]], optional): The map with the toolbar's buttons
            refresh (Callable[[int, int], int], optional): A lambda function that refreshes a toolbar
        """
        super(ToolBar, self).__init__(parent)

        if self.parent and hasattr(self.parent, 'font'):
            # Apply font from the main window to the dialog
            self.setFont(self.parent.font())

        self.logger = logging.getLogger('toolbar')

        self.logging = AppConfig.get_logging()
        self.debug = AppConfig.get_debug()

        self.actions = actions if actions else {}
        self.buttons = buttons if buttons else {}
        self.refresh = refresh

        self.settings = Settings()

        self.theme_helper = ThemeHelper()

        # Default language setup, change to settings value to modify it via UI
        self.lexemes = Lexemes(self.settings.app_language, default_scope='toolbar')

        # Sometimes (when weights changed) the values could be reset like this:
        # self.settings.toolbar_icons = None

        if self.settings.toolbar_icons is None or self.settings.toolbar_icons == 0:
            """
            Default bit-mask is 131070 when all icons are checked.
            """
            self.settings.toolbar_icons = 131070
            if self.debug:
                self.logger.info('Starting with default icons mask "%d"' % self.settings.toolbar_icons)

        self.search_input = None  # type: Union[QLineEdit, None]
        self.search_match_case = None  # type: Union[QCheckBox, None]
        self.search_input = None  # type: Union[QLineEdit, None]
        self.btn_search_clear = None  # type: Union[QPushButton, None]
        self.btn_search_prev = None  # type: Union[QPushButton, None]
        self.btn_search_next = None  # type: Union[QPushButton, None]
        self.search_match_case = None  # type: Union[QCheckBox, None]

        self.toolbar_save_button = None  # type: Union[QToolButton, None]
        self.toolbar_edit_button = None  # type: Union[QToolButton, None]

        self.setMovable(False)

        self.init_ui()

    def init_ui(self):
        prev_type = None
        for icon in self.actions:
            if icon['type'] == 'action':
                # Check visible icon is checked in settings
                if not (self.settings.toolbar_icons & pow(2, icon['weight'])):
                    # Skip if not checked in settings
                    continue
                # If the icon has an active state check, check it here
                if ('active_state_check' in icon
                        and callable(icon['active_state_check'])
                        and not icon['active_state_check']()):
                    # Skip if the icon isn't active
                    continue
                # Create and append toolbar icon
                self.append_toolbar_icon(icon)
            elif icon['type'] == 'delimiter' and prev_type != 'delimiter':
                self.addSeparator()
            # Save prev type
            prev_type = icon['type']

        central_spacer = QWidget(self)
        central_spacer.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.addWidget(central_spacer)

        search_label = QLabel(self)
        # Apply font from the main window to the widget
        search_label.sizeHint()
        search_label.setText(self.lexemes.get('search_input_label'))
        search_label.setObjectName('search_input_label')  # To differentiate it at styles file
        self.addWidget(search_label)

        self.search_input = QLineEdit(self)
        self.search_input.setObjectName('search_input')
        self.search_input.sizeHint()
        self.search_input.setReadOnly(False)
        self.search_input.setMaxLength(128)
        self.search_input.setPlaceholderText(self.lexemes.get('search_input_placeholder_text'))
        self.search_input.setAccessibleDescription(
            self.lexemes.get('search_input_accessible_description'))

        self.addWidget(self.search_input)

        for button in self.buttons:
            if button['type'] == 'action':
                # Create and append toolbar button
                self.append_toolbar_button(button)

        search_match_case_label = QLabel(self)
        search_match_case_label.sizeHint()
        search_match_case_label.setText(self.lexemes.get('search_case_sensitive'))
        search_match_case_label.setObjectName("search_case_sensitive")  # To differentiate it at styles file
        self.addWidget(search_match_case_label)

        self.search_match_case = QCheckBox(self)
        self.search_match_case.sizeHint()
        self.search_match_case.setToolTip(self.lexemes.get('search_match_case_tooltip'))
        self.search_match_case.setCheckState(Qt.CheckState.Unchecked)
        self.addWidget(self.search_match_case)

        """
        It can be done by setting inline styles, but this is not what can be convenient for customisation,
        hence the theme:
        """
        # self.toolbar.setStyleSheet("""QCheckBox {
        #    margin-right: 5px;
        # }""")
        self.setStyleSheet(self.theme_helper.get_css('toolbar'))

    def append_toolbar_icon(self, conf):
        """
        Helper to create, add to the toolbar and return button with an icon.
        """
        # Theme icon with a fallback to a system one
        system_icon = conf['system_icon'] if 'system_icon' in conf else None
        theme_icon = conf['theme_icon'] if 'theme_icon' in conf else None
        theme_icon_color = QColor(conf['color']) if 'color' in conf \
            else self.theme_helper.get_color('toolbar_icon_color_default')
        icon = self.theme_helper.get_icon(theme_icon=theme_icon, system_icon=system_icon, color=theme_icon_color)

        # Button's action
        label = conf['label'] if 'label' in conf else ''
        icon_action = QAction(icon, label, self)
        action = conf['action'] if 'action' in conf else None  # Action triggered on click
        if action is not None:
            icon_action.triggered.connect(action)

        # Toolbar button itself
        icon_button = QToolButton(self)
        if 'name' in conf:
            icon_button.setObjectName(conf['name'])
        icon_button.setToolTip(label)
        # icon_button.setIconSize(QSize(16, 16));
        icon_button.setDefaultAction(icon_action)
        # icon_action.setChecked(True)
        if 'accessible_name' in conf:
            icon_button.setAccessibleName(conf['accessible_name'])

        # Add the button to the toolbar
        self.addWidget(icon_button)

        # Add internal variable to access the icon later, say for state toggle
        if 'var_name' in conf:
            if self.debug and hasattr(self, conf['var_name']):
                self.logger.debug('Variable "%s" is already set! Re-writing it...' % conf['var_name'])
            setattr(self, conf['var_name'], icon_button)  # type: QToolButton
        # If the icon has a switched off check, check it here
        if ('switched_off_check' in conf
                and callable(conf['switched_off_check'])
                and conf['switched_off_check']()):
            # Switch the icon off
            icon_button.setDisabled(True)

    def append_toolbar_button(self, conf):
        """
        Helper to create and return toolbar push button.
        """
        # Theme icon with a fallback to a system one
        system_icon = conf['system_icon'] if 'system_icon' in conf else None
        theme_icon = conf['theme_icon'] if 'theme_icon' in conf else None
        theme_icon_color = QColor(conf['color']) if 'color' in conf \
            else self.theme_helper.get_color('toolbar_icon_color_default')
        # Theme icon with a fallback to a system one
        icon = self.theme_helper.get_icon(theme_icon=theme_icon, system_icon=system_icon, color=theme_icon_color)

        # Toolbar button with an icon
        icon_button = QPushButton(self)
        icon_button.setFont(self.font())
        # Set the size (height) similar to search input field height
        if hasattr(self, 'search_input'):
            # Adjust size to maintain ratio
            size_hint = self.search_input.sizeHint()  # As a real height hint check the line edit height
            icon_button.setFixedSize(QSize(size_hint.height(), size_hint.height()))
            icon_height = int(size_hint.height() * 0.7)
            icon_button.setIconSize(QSize(icon_height, icon_height))
        icon_button.setIcon(icon)

        action = conf['action'] if 'action' in conf else None  # Action when clicked
        if action is not None:
            icon_button.clicked.connect(action)
        if 'accessible_name' in conf:
            icon_button.setAccessibleName(conf['accessible_name'])
        if 'tooltip' in conf:
            icon_button.setToolTip(conf['tooltip'])
        if 'default' in conf:
            icon_button.setDefault(conf['default'])
        if 'enabled' in conf:
            icon_button.setEnabled(conf['enabled'])

        # Add the button to the toolbar
        self.addWidget(icon_button)

        if 'var_name' in conf:
            if self.debug and hasattr(self, conf['var_name']):
                self.logger.debug('Variable "%s" is already set! Re-writing it...' % conf['var_name'])
            setattr(self, conf['var_name'], icon_button)  # type: QPushButton

    def contextMenuEvent(self, event):
        """
        Render context menu event upon the toolbar right mouse click
        """
        current_action = self.actionAt(event.pos())
        if current_action is None:
            return

        if self.actions is None:
            return

        menu = QMenu(self)

        _weights = 0
        for index, label in enumerate(self.actions, 1):
            if 'type' not in label or label['type'] != 'action':
                menu.addSeparator()
                continue

            # Get settings weight of the item
            settings_weight = pow(2, label['weight'])

            if _weights & settings_weight:
                # The item is already added or the active state item that shouldn't be duplicated
                continue

            button = QAction(label['label'], self)
            button.setFont(self.font())

            slot = lambda checked, i=label['weight']: self.toolbar_menu_item(checked, i)
            # Check button.toggled.connect()
            button.triggered[bool].connect(slot)
            button.setCheckable(True)
            # Check the item is in the settings weights
            button.setChecked(self.settings.toolbar_icons & settings_weight)

            menu.addAction(button)

            self.buttons[index] = button

            # Collect items already added to the context menu
            _weights |= settings_weight

        # PoC remove element from the toolbar
        # delete_action  = menu.addAction("Hide")

        menu.exec(event.globalPos())

        # PoC Remove element from the toolbar
        # if action == delete_action:
        #    self.removeAction(current_action)

    def toolbar_menu_item(self, checked: bool, index: int) -> None:
        """
        Draw context menu item with a checkbox

        Args:
            checked (bool): Checked or not
            index (int): Index in a toolbar menu mapping
        """
        pi = pow(2, index)
        if self.debug:
            self.logger.info('checked:{} index:{} pi:{}' . format(checked, index, pi))
        if index in self.buttons:
            self.buttons[index].setChecked(checked)
        if checked:
            self.settings.toolbar_icons |= pi
        else:
            self.settings.toolbar_icons ^= pi
        # Re-draw toolbar with a callback
        if callable(self.refresh):
            self.refresh()

    # PoC Remove element from the toolbar
    # def removeAction(self, action):
    #    """
    #    Override removeAction() method upon QWidgetAction
    #
    #    Args:
    #        action (QAction):
    #    """
    #    super(ToolBar, self).removeAction(action)
    #
    #    if self.debug:
    #        self.logger.info('Remove element action "%s"' % action)



import os
import logging
import importlib.util

from typing import Union, Dict


class Lexemes:
    def __init__(self, language: str = 'en', default_scope: str = 'common'):
        if not language or language in ('.', '..'):
            language = 'en'

        self.language = language
        self._default_scope = default_scope

        # Lexeme storage
        self.lexemes = {}

        self.logger = logging.getLogger('lexemes')

        self.lexemes = self.load_lexemes()

    def get_lexemes_dir(self):
        return os.path.join(os.path.dirname(__file__), self.language)

    def load_lexemes(self) -> Dict[str, Dict]:
        # Get the directory path of lexemes relative to this file
        lexemes_lang_dir = self.get_lexemes_dir()

        lexeme_files = []
        if os.path.isdir(lexemes_lang_dir):
            for file_name in os.listdir(lexemes_lang_dir):
                # Load all .py files from the dir excluding __init__.py file if exists
                if file_name.endswith('.py') and file_name != '__init__.py':
                    lexeme_files.append(os.path.join(lexemes_lang_dir, file_name))

        # Import and collect lexemes from each lexeme file
        lexemes = {}
        for file_path in lexeme_files:
            scope_name = os.path.splitext(os.path.basename(file_path))[0]
            if scope_name not in lexemes:
                lexemes[scope_name] = {}
            spec = importlib.util.spec_from_file_location(scope_name, file_path)
            lexemes_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(lexemes_module)
            if hasattr(lexemes_module, 'lexemes'):
                lexemes[scope_name].update(lexemes_module.lexemes)

        return lexemes

    def set_language(self, language: str) -> None:
        """
        Set language and re-load all corresponded lexemes
        @param language: string key of the language to load
        @return: None
        """
        self.language = language
        self.load_lexemes()

    def get(self, name, scope: str = None, **kwargs) -> Union[str, None]:
        """
        Get lexeme by its name and scope (optional; default 'common')
        @param name: lexeme's name key
        @param scope: lexeme's scope, e.g. common, encrypt, etc. If not set default_scope is in use.
        @return: either string value or None if lexeme's scope is not found
        """
        if not scope:
            # Use this method to get default scope as it may contain extra logic
            scope = self.get_default_scope()

        lexeme = self.lexemes[scope].get(name, '') if scope in self.lexemes else None
        if isinstance(lexeme, bytes):
            lexeme = lexeme.decode('utf-8')

        if kwargs:
            # Process placeholders if set, e.g. name="abc" in a string like "... {name} ..." become "... abc ..."
            try:
                lexeme = lexeme.format(**kwargs)
            except KeyError:
                self.logger.warning('Lexeme "%s" params [%s] not found for replacement'
                                    % (lexeme, ', '.join(kwargs.keys())))
        return lexeme

    def get_all(self):
        # Return all lexemes set
        return self.lexemes

    def get_default_scope(self):
        # Default scope if set
        scope = self._default_scope

        # If scope is not found in loaded lexemes
        if scope not in self.lexemes:
            # The very last fallback option
            scope = 'common'

        return scope

    def get_full_key(self, name, scope: str = None):
        if not scope:
            # Use this method to get default scope as it may contain extra logic
            scope = self.get_default_scope()

        return f"{scope}_{name}"



class EncPassword:
    """
    Password object class to store password related data.
    """

    def __init__(self):
        self._password = None
        self._hint = None

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, value):
        self._password = value

    @property
    def hint(self):
        return self._hint

    @hint.setter
    def hint(self, value):
        self._hint = value

    def is_valid(self):
        if self._password is None:
            return False
        if len(self._password) < 8:
            return False
        return True



# English lexemes settings_dialog.py
lexemes = {
    # Settings dialog
    "window_title": "Settings",

    "button_close": "Close",

    "tab_general": "General",
    "tab_editor_config": "Editor",
    "tab_viewer_config": "Viewer",
    "tab_ai_config": "AI Config",

    "general_app_config_label": "App config",
    "general_app_language_label": "Language",
    "general_app_language_combo_placeholder_text": "Choose a language",
    "general_app_language_combo_accessible_description": "App's interface language",
    "general_app_theme_label": "Theme",
    "general_app_theme_combo_placeholder_text": "Choose a theme",
    "general_app_theme_combo_accessible_description": "App's interface theme",
    "general_app_main_menu_label": "Main Menu",
    "general_app_main_menu_checkbox": "Show Main Menu",
    "general_app_main_menu_checkbox_accessible_description": "Display the app's main dropdown menu",
    "general_app_font_size_label": "Font Size: {size}pt",
    "general_app_font_size_slider_accessible_description": "Adjust the app's global font size",

    "general_statusbar_label": "Status Bar",
    "general_statusbar_show_global_cursor_position_checkbox": "Show Global Cursor Position",
    "general_statusbar_show_global_cursor_position_checkbox_accessible_description":
        "Display the global cursor position in the status bar",

    "editor_config_label": "Editor Configuration",
    "editor_config_show_line_numbers_checkbox": "Show Line Numbers",
    "editor_config_show_line_numbers_checkbox_accessible_description": "Display line numbers in the editor",

    "viewer_config_label": "Viewer Configuration",
    "viewer_config_process_emojis_checkbox": "Convert Text Emojis to Graphics",
    "viewer_config_process_emojis_checkbox_accessible_description": "Convert text emojis to graphical representations",
    "viewer_config_highlight_todos_checkbox": "Highlight TODOs",
    "viewer_config_highlight_todos_checkbox_accessible_description": "Emphasize TODO tags within the text",
    "viewer_config_open_link_confirmation_checkbox": "Require Confirmation to Open Links",
    "viewer_config_open_link_confirmation_checkbox_accessible_description":
        "Ask for confirmation before opening links",
    "viewer_config_save_resources_checkbox": "Auto-save external images to disk",
    "viewer_config_save_resources_checkbox_accessible_description":
        "Automatically saves copies of external images to disk for offline access.",

    "ai_config_openai_api_label": "OpenAI API",
    "ai_config_openai_api_url_input_placeholder_text": "API URL",
    "ai_config_openai_api_url_input_accessible_description": "OpenAI API URL",
    "ai_config_openai_api_key_input_placeholder_text": "API key",
    "ai_config_openai_api_key_input_accessible_description": "OpenAI API key",
    "ai_config_openai_api_supported_models_label": "Supported models",
    "ai_config_ai_model_names_combo_placeholder_text": "Choose model",
    "ai_config_ai_model_names_combo_accessible_description": "Supported models to choose",

    "ai_config_base_label": "Base Parameters",
    "ai_config_base_system_prompt_label": "System Prompt",
    "ai_config_base_system_prompt_edit_placeholder_text": "Base system prompt that precedes each request",
    "ai_config_base_system_prompt_edit_accessible_description":
        "Base system prompt that precedes each request. This is plain text.",
    "ai_config_base_response_max_tokens_label": "Maximum Response Tokens",
    "ai_config_base_response_max_tokens_input_accessible_description": "Maximum number of tokens to receive in response",
}


# tests/test_settings.py
from notolog.settings import Settings

import pytest

class TestSettings():

    def test_settings_defaults(self):
        obj = Settings()
        obj.clear()
        assert obj.ui_width == 0
        assert obj.ui_height == 0
        assert obj.ui_pos_x == 0
        assert obj.ui_pos_y == 0
        assert obj.file_path is ''
        assert obj.toolbar_icons is 0

    def test_settings_setter_getter(self):
        settings_a = Settings()
        # Reset to test
        settings_a.clear()

        # Set values in A object
        settings_a.ui_width = 11
        settings_a.ui_height = 22
        settings_a.ui_pos_x = 33
        settings_a.ui_pos_y = 44
        settings_a.file_path = 'test/path/file.txt'
        settings_a.toolbar_icons = 8

        settings_b = Settings()

        # Assert values are stored correctly
        assert settings_b.ui_width == 11
        assert settings_b.ui_height == 22
        assert settings_b.ui_pos_x == 33
        assert settings_b.ui_pos_y == 44
        assert settings_b.file_path == 'test/path/file.txt'
        assert settings_b.toolbar_icons is 8

        # Reset to restore
        settings_b.clear()



from PySide6.QtWidgets import QPushButton

from qasync import QEventLoop, QApplication, asyncSlot
import asyncio

from notolog.notolog_editor import NotologEditor
from notolog.image_downloader import ImageDownloader

import os
import sys
import pytest


class TestQtAsync:

    @pytest.fixture(autouse=True, scope="class")  # Auto-use as this fixture should be used for all tests
    def qt_app(self):
        # Force Qt style override to "Fusion" to avoid issues on environments where "kvantum" is not available
        os.environ["QT_STYLE_OVERRIDE"] = "Fusion"
        # Create the QApplication instance for the tests
        app = QApplication(sys.argv)

        # Initialize the QEventLoop for asyncio integration with Qt
        loop = QEventLoop(app)
        asyncio.set_event_loop(loop)

        yield app

        # Properly close the QApplication after each test
        app.quit()

    @pytest.fixture
    def main_window(self, mocker, qt_app):
        # Avoid downloading resources during tests
        mocker.patch.object(NotologEditor, 'is_resource_attached', return_value=True)
        mocker.patch.object(ImageDownloader, 'download_resource_in_queue', return_value=None)

        # Fixture to create and return main window instance
        window = NotologEditor(screen=qt_app.screens()[0])
        yield window

    @pytest.fixture
    async def cleanup_tasks(self):
        yield
        # Cancel and clean up all pending asyncio tasks
        for task in asyncio.all_tasks():
            if task is not asyncio.current_task():
                task.cancel()
                try:
                    await task
                except asyncio.CancelledError:
                    pass

    @asyncSlot()
    async def on_click(self):
        # Simulate some async operation by sleeping
        await asyncio.sleep(0.1)
        # Append "clicked" to the class-level result list
        self.result.append("clicked")

    @pytest.mark.asyncio
    async def test_ui_interaction(self, mocker, qt_app, main_window, cleanup_tasks):
        # Reset result list before each test if needed
        self.result = []

        button = QPushButton("Test Button")
        button.clicked.connect(lambda:
                               asyncio.ensure_future(self.on_click()))

        # Simulate a button click
        button.click()

        # Immediately after click, result should still be empty before the sleep allows on_click to proceed
        assert self.result == []
        # Wait for the sleep in on_click to finish
        await asyncio.sleep(0.25)  # Ensure this is long enough for the on_click task to complete
        # Now, result should have "clicked"
        assert self.result == ["clicked"]

        mock_rehighlight = mocker.patch.object(main_window, 'rehighlight_task_callback')

        # Check local task pool is empty
        assert len(main_window.rehighlight_tasks) == 0

        # Queue re-highlight tasks
        task0 = await main_window.rehighlight_in_queue()
        mock_rehighlight.assert_called_once_with(task0)
        # One more
        task1 = await main_window.rehighlight_in_queue()
        mock_rehighlight.assert_called_with(task1)
        # Another one more
        task2 = await main_window.rehighlight_in_queue()
        mock_rehighlight.assert_called_with(task2)
        # Add more tasks
        await main_window.rehighlight_in_queue()
        await main_window.rehighlight_in_queue()
        await main_window.rehighlight_in_queue()
        await main_window.rehighlight_in_queue()
        await main_window.rehighlight_in_queue()
        # Max count in a short period of time is 3
        assert mock_rehighlight.call_count == 3

        # Checking of local task pool doesn't work properly if run from IDE, but works from CLI
        assert len(main_window.rehighlight_tasks) == 3

        # The cleanup_tasks fixture is used implicitly; no need to assert its presence
        # assert cleanup_tasks



from .enum_base import EnumBase

assert hasattr(EnumBase, 'default'), "Check default() method is implemented in the base class"


class Colors(EnumBase):

    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        """
        Override _generate_next_value_ to allow for custom attributes on enum members (mostly for default).

        'name' is the name of the Enum's member, which will be 'A', 'B', 'C', etc.
        This method returns the name itself, an 'extended' marker (False), and a 'default' marker (False).

        Extended means extended color palette.
        """
        return name, False, False

    # Red Spectrum
    CRIMSON = ("Crimson", True)
    DARKRED = ("DarkRed", True)
    FIREBRICK = ("Firebrick", True)
    INDIANRED = ("IndianRed", True)
    LIGHTCORAL = ("LightCoral", True)
    MAROON = ("Maroon", False)
    RED = ("Red", False)
    SALMON = ("Salmon", True)
    DARKSALMON = ("DarkSalmon", True)
    TOMATO = ("Tomato", True)

    # Orange Spectrum
    DARKORANGE = ("DarkOrange", True)
    ORANGE = ("Orange", True)
    CORAL = ("Coral", True)
    ORANGERED = ("OrangeRed", True)

    # Yellow Spectrum
    GOLD = ("Gold", True)
    YELLOW = ("Yellow", False)
    LIGHTGOLDENRODYELLOW = ("LightGoldenrodYellow", True)
    PALEGOLDENROD = ("PaleGoldenrod", True)
    GOLDENROD = ("Goldenrod", True)
    DARKGOLDENROD = ("DarkGoldenrod", True)

    # Green Spectrum
    GREENYELLOW = ("GreenYellow", True)
    CHARTREUSE = ("Chartreuse", True)
    LAWNGREEN = ("LawnGreen", True)
    LIME = ("Lime", False)
    LIMEGREEN = ("LimeGreen", True)
    PALEGREEN = ("PaleGreen", True)
    LIGHTGREEN = ("LightGreen", True)
    MEDIUMSPRINGGREEN = ("MediumSpringGreen", True)
    SPRINGGREEN = ("SpringGreen", True)
    MEDIUMSEAGREEN = ("MediumSeaGreen", True)
    SEAGREEN = ("SeaGreen", True)
    FORESTGREEN = ("ForestGreen", True)
    GREEN = ("Green", False)
    DARKGREEN = ("DarkGreen", True)
    YELLOWGREEN = ("YellowGreen", True)
    OLIVEDRAB = ("OliveDrab", True)
    OLIVE = ("Olive", False)

    # Cyan to Blue Spectrum
    MEDIUMAQUAMARINE = ("MediumAquamarine", True)
    AQUAMARINE = ("Aquamarine", True)
    TURQUOISE = ("Turquoise", True)
    MEDIUMTURQUOISE = ("MediumTurquoise", True)
    DARKTURQUOISE = ("DarkTurquoise", True)
    LIGHTSEAGREEN = ("LightSeaGreen", True)
    CADETBLUE = ("CadetBlue", True)
    DARKCYAN = ("DarkCyan", True)
    TEAL = ("Teal", False)
    CYAN = ("Cyan", True)
    LIGHTCYAN = ("LightCyan", True)
    AQUA = ("Aqua", False)
    PALETURQUOISE = ("PaleTurquoise", True)
    AZURE = ("Azure", True)
    ALICEBLUE = ("AliceBlue", True)

    # Blue Spectrum
    NAVY = ("Navy", False)
    DARKBLUE = ("DarkBlue", True)
    MEDIUMBLUE = ("MediumBlue", True)
    BLUE = ("Blue", False)
    MIDNIGHTBLUE = ("MidnightBlue", True)
    ROYALBLUE = ("RoyalBlue", True)
    STEELBLUE = ("SteelBlue", True)
    DODGERBLUE = ("DodgerBlue", True)
    DEEPSKYBLUE = ("DeepSkyBlue", True)
    CORNFLOWERBLUE = ("CornflowerBlue", True)
    SKYBLUE = ("SkyBlue", True)
    LIGHTSKYBLUE = ("LightSkyBlue", True)
    SLATEBLUE = ("SlateBlue", True)
    DARKSLATEBLUE = ("DarkSlateBlue", True)
    MEDIUMSLATEBLUE = ("MediumSlateBlue", True)

    # Violet and Magenta Spectrum
    LAVENDER = ("Lavender", True)
    BLUEVIOLET = ("BlueViolet", True)
    DARKVIOLET = ("DarkViolet", True)
    PURPLE = ("Purple", False)
    MEDIUMPURPLE = ("MediumPurple", True)
    INDIGO = ("Indigo", True)
    DARKORCHID = ("DarkOrchid", True)
    DARKMAGENTA = ("DarkMagenta", True)
    MAGENTA = ("Magenta", True)
    ORCHID = ("Orchid", True)
    VIOLET = ("Violet", True)
    PLUM = ("Plum", True)
    THISTLE = ("Thistle", True)
    LAVENDERBLUSH = ("LavenderBlush", True)
    MEDIUMVIOLETRED = ("MediumVioletRed", True)
    PALEVIOLETRED = ("PaleVioletRed", True)

    # Neutral and Light Spectrum
    GAINSBORO = ("Gainsboro", True)
    LIGHTGREY = ("LightGrey", True)
    SILVER = ("Silver", False)
    DARKGRAY = ("DarkGray", True)
    GRAY = ("Gray", False)
    DIMGRAY = ("DimGray", True)
    LIGHTSLATEGRAY = ("LightSlateGray", True)
    SLATEGRAY = ("SlateGray", True)
    DARKSLATEGRAY = ("DarkSlateGray", True)
    BLACK = ("Black", False, True)
    WHITE = ("White", False)
    SNOW = ("Snow", True)
    HONEYDEW = ("Honeydew", True)
    MINTCREAM = ("MintCream", True)
    GHOSTWHITE = ("GhostWhite", True)
    WHITESMOKE = ("WhiteSmoke", True)
    SEASHELL = ("Seashell", True)
    BEIGE = ("Beige", True)
    OLDLACE = ("OldLace", True)
    FLORALWHITE = ("FloralWhite", True)
    IVORY = ("Ivory", True)
    ANTIQUEWHITE = ("AntiqueWhite", True)
    LINEN = ("Linen", True)
    MISTYROSE = ("MistyRose", True)
    PAPAYAWHIP = ("PapayaWhip", True)
    BLANCHEDALMOND = ("BlanchedAlmond", True)
    BISQUE = ("Bisque", True)
    PEACHPUFF = ("PeachPuff", True)
    NAVAJOWHITE = ("NavajoWhite", True)
    MOCCASIN = ("Moccasin", True)
    CORNSILK = ("CornSilk", True)
    LEMONCHIFFON = ("LemonChiffon", True)
    KHAKI = ("Khaki", True)
    WHEAT = ("Wheat", True)

    def __init__(self, value, is_extended=False, is_default=False):
        # If the value is a tuple, unpack it.
        if isinstance(value, tuple):
            self._value_, self.is_extended, self.is_default = value
        else:
            self._value_ = value
            self.is_extended = is_extended
            self.is_default = is_default

    @classmethod
    def extended_members(cls):
        return [member for member in cls if member.is_extended]

    @classmethod
    def is_extended(cls, member):
        return member in cls.extended_members()



from .enum_base import EnumBase

assert hasattr(EnumBase, 'default'), "Check default() method is implemented in the base class"


class Themes(EnumBase):

    DEFAULT = ("Default", True)  # The second item in the tuple marks this as the default
    CALLIGRAPHY = "Calligraphy"
    NOIR_DARK = "Noir Dark"
    STRAWBERRY = "Strawberry"

    def __init__(self, value, is_default=False):
        self._value_ = value
        self.is_default = is_default



# German lexemes settings_dialog.py
lexemes = {
    # Einstellungen
    "window_title": "Einstellungen",

    "button_close": "Schließen",

    "tab_general": "Allgemein",
    "tab_editor_config": "Editor",
    "tab_viewer_config": "Betrachter",
    "tab_ai_config": "KI-Konfiguration",

    "general_app_config_label": "App-Konfiguration",
    "general_app_language_label": "Sprache",
    "general_app_language_combo_placeholder_text": "Sprache wählen",
    "general_app_language_combo_accessible_description": "Sprache der App-Oberfläche",
    "general_app_theme_label": "Thema",
    "general_app_theme_combo_placeholder_text": "Thema wählen",
    "general_app_theme_combo_accessible_description": "Thema der App-Oberfläche",
    "general_app_main_menu_label": "Hauptmenü",
    "general_app_main_menu_checkbox": "Hauptmenü anzeigen",
    "general_app_main_menu_checkbox_accessible_description": "Das Haupt-Dropdown-Menü der App anzeigen",
    "general_app_font_size_label": "Schriftgröße: {size}pt",
    "general_app_font_size_slider_accessible_description": "Globale Schriftgröße der App anpassen",

    "general_statusbar_label": "Statusleiste",
    "general_statusbar_show_global_cursor_position_checkbox": "Globale Cursorposition anzeigen",
    "general_statusbar_show_global_cursor_position_checkbox_accessible_description":
        "Globale Cursorposition in der Statusleiste anzeigen",

    "editor_config_label": "Editor-Konfiguration",
    "editor_config_show_line_numbers_checkbox": "Zeilennummern anzeigen",
    "editor_config_show_line_numbers_checkbox_accessible_description": "Zeilennummern im Editor anzeigen",

    "viewer_config_label": "Viewer-Konfiguration",
    "viewer_config_process_emojis_checkbox": "Text-Emojis in Grafiken umwandeln",
    "viewer_config_process_emojis_checkbox_accessible_description": "Text-Emojis in grafische Darstellungen umwandeln",
    "viewer_config_highlight_todos_checkbox": "TODOs hervorheben",
    "viewer_config_highlight_todos_checkbox_accessible_description": "TODOs im Text hervorheben",
    "viewer_config_open_link_confirmation_checkbox": "Bestätigung vor dem Öffnen von Links erforderlich",
    "viewer_config_open_link_confirmation_checkbox_accessible_description":
        "Vor dem Öffnen von Links eine Bestätigung anfordern",
    "viewer_config_save_resources_checkbox": "Externe Bilder automatisch auf Festplatte speichern",
    "viewer_config_save_resources_checkbox_accessible_description":
        "Speichert automatisch Kopien von externen Bildern auf der Festplatte für den Offline-Zugriff.",

    "ai_config_openai_api_label": "OpenAI-API",
    "ai_config_openai_api_url_input_placeholder_text": "API-URL",
    "ai_config_openai_api_url_input_accessible_description": "URL der OpenAI-API",
    "ai_config_openai_api_key_input_placeholder_text": "API-Schlüssel",
    "ai_config_openai_api_key_input_accessible_description": "API-Schlüssel von OpenAI",
    "ai_config_openai_api_supported_models_label": "Unterstützte Modelle",
    "ai_config_ai_model_names_combo_placeholder_text": "Modell wählen",
    "ai_config_ai_model_names_combo_accessible_description": "Zur Auswahl stehende unterstützte Modelle",

    "ai_config_base_label": "Grundparameter",
    "ai_config_base_system_prompt_label": "Systemaufforderung",
    "ai_config_base_system_prompt_edit_placeholder_text": "Grundsystemaufforderung, die jeder Anfrage vorausgeht",
    "ai_config_base_system_prompt_edit_accessible_description":
        "Grundsystemaufforderung, die jeder Anfrage vorausgeht. Reiner Text.",
    "ai_config_base_response_max_tokens_label": "Maximale Antwort-Token",
    "ai_config_base_response_max_tokens_input_accessible_description":
        "Maximale Anzahl von Token, die in der Antwort empfangen werden",
}


from PySide6.QtCore import Qt
from PySide6.QtWidgets import QDialog, QLabel, QVBoxLayout, QDialogButtonBox, QSizePolicy
from PySide6.QtGui import QFontMetrics

from . import AppConfig

import logging


class CommonDialog(QDialog):

    debug = AppConfig.get_debug()

    def __init__(self, title=str, text=str, callback=None, parent=None):
        super().__init__(parent)

        self.parent = parent

        if self.parent and hasattr(self.parent, 'font'):
            # Apply font from the main window to the dialog
            self.setFont(self.parent.font())

        self.logger = logging.getLogger('common_dialog')

        self.setWindowTitle(str(title))
        self.setObjectName('common_dialog')

        # Set dialog size derived from the main window size
        """
        main_window_size = parent.size()
        # Will be adjusted with size hint later at resizeEvent()
        dialog_width = int(main_window_size.width() * 0.33)
        dialog_height = int(main_window_size.height() * 0.33)
        self.resize(dialog_width, dialog_height)
        """
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        if self.sizeHint().isValid():
            self.setMinimumSize(self.sizeHint())

        label = QLabel()
        label.setObjectName('common_label')
        label.setMargin(10)  # Set vertical padding for the dialog label
        label.setText(str(text))
        label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        label.adjustSize()
        """
        Other examples of the widget geometry's adjustments:
        label.move(128, 64)
        label.setGeometry(QStyle.alignedRect(
            Qt.LayoutDirection.LeftToRight,
            Qt.AlignmentFlag.AlignCenter,
            dialog.size(),
            dialog.geometry()))
        """

        button_box = QDialogButtonBox()
        button_box.setObjectName('button_box')
        button_box.setOrientation(Qt.Orientation.Horizontal)
        button_box.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Yes)
        button_box.setCenterButtons(True)

        # not QVBoxLayout(self) to avoid: 'Attempting to add QLayout "" to ... "", which already has a layout
        layout = QVBoxLayout()
        layout.addWidget(label)
        layout.addWidget(button_box)
        self.setLayout(layout)

        self.adjust_minimum_width(str(title))

        if callable(callback):
            button_box.accepted.connect(lambda: callback(self.reject))
        else:
            if self.debug:
                self.logger.warning('No callback method provided for the dialog!')
            button_box.accepted.connect(self.close)

        button_box.rejected.connect(self.reject)

    def resizeEvent(self, event) -> None:
        """
        To make dialog stretch if the content size is longer than expected
        @param event: QEvent
        @return: None
        """
        # Get the preferred size of the dialog's content
        preferred_size = self.layout().sizeHint()
        # Adjust the dialog's size based on the preferred size then
        if preferred_size.isValid():
            self.resize(preferred_size)

    def adjust_minimum_width(self, title: str) -> None:
        # Calculate width required for the title text
        font_metrics = QFontMetrics(self.font())
        title_width = font_metrics.horizontalAdvance(title)

        # Set minimum width for the dialog
        self.setMinimumWidth(title_width + 150)  # Add some padding



import os
import tomli

toml_base_app_config = """
[app]
name = "Notolog"
version = "0.9.1b4"
license = "MIT"
date = "2024"
website = "https://notolog.app"
repository = "https://github.com/notolog/notolog-editor"
pypi = "https://pypi.org/project/notolog"
author = "Vadim Bakhrenkov"

[repository]

    [repository.github]
    username = "notolog"
    project = "notolog-editor"
    release_url = "https://api.github.com/repos/notolog/notolog-editor/releases/latest"
    bug_report_url = "https://github.com/notolog/notolog-editor/issues"
"""

# This is a fallback option if app_config.toml is not found
toml_app_config = """
[font]
base_size = 12
min_size = 5
max_size = 42

[logger]
logging = true
debug = false
"""


class AppConfig:
    """
    App configuration to share between app's objects.
    """

    base_app_config = tomli.loads(toml_base_app_config)

    # Get this file dir path
    script_dir = os.path.dirname(os.path.realpath(__file__))
    # Get parent dir path
    parent_dir = os.path.dirname(script_dir)
    # Get app_config.toml file path which is located at parent dir
    toml_file_path = os.path.join(parent_dir, 'app_config.toml')

    if not os.path.exists(toml_file_path):
        # Backup option
        app_config = tomli.loads(toml_app_config)
    else:
        # Read actual app_config.toml
        with open(toml_file_path, 'r') as config_file:
            app_config = tomli.loads(config_file.read())

    # Font settings
    _font_size = app_config['font']['base_size']  # Base value
    _prev_font_size = app_config['font']['base_size']  # Previous value

    # For pytest to allow override some params
    _test_mode = False

    """
    Constant values with getters.
    """

    @classmethod
    def get_app_name(cls) -> str:
        return cls.base_app_config['app']['name']

    @classmethod
    def get_app_version(cls) -> str:
        return cls.base_app_config['app']['version']

    @classmethod
    def get_app_license(cls) -> str:
        return cls.base_app_config['app']['license']

    @classmethod
    def get_app_date(cls) -> str:
        return cls.base_app_config['app']['date']

    @classmethod
    def get_app_website(cls) -> str:
        return cls.base_app_config['app']['website']

    @classmethod
    def get_app_repository(cls) -> str:
        return cls.base_app_config['app']['repository']

    @classmethod
    def get_app_pypi(cls) -> str:
        return cls.base_app_config['app']['pypi']

    @classmethod
    def get_app_author(cls) -> str:
        return cls.base_app_config['app']['author']

    @classmethod
    def get_repository_github_username(cls) -> str:
        return cls.base_app_config['repository']['github']['username']

    @classmethod
    def get_repository_github_project(cls) -> str:
        return cls.base_app_config['repository']['github']['username']

    @classmethod
    def get_repository_github_release_url(cls) -> str:
        return cls.base_app_config['repository']['github']['release_url']

    @classmethod
    def get_repository_github_bug_report_url(cls) -> str:
        return cls.base_app_config['repository']['github']['bug_report_url']

    @classmethod
    def get_font_base_size(cls) -> int:
        return cls.app_config['font']['base_size']

    """
    Methods with setter to override default value.
    """

    @classmethod
    def set_logging(cls, value) -> None:
        cls.app_config['logger']['logging'] = value

    @classmethod
    def get_logging(cls) -> bool:
        return cls.app_config['logger']['logging']

    @classmethod
    def set_debug(cls, value) -> None:
        cls.app_config['logger']['debug'] = value

    @classmethod
    def get_debug(cls) -> bool:
        return cls.app_config['logger']['debug']

    @classmethod
    def set_font_min_size(cls, value) -> None:
        cls.app_config['font']['min_size'] = value

    @classmethod
    def get_font_min_size(cls) -> int:
        return cls.app_config['font']['min_size']

    @classmethod
    def set_font_max_size(cls, value) -> None:
        cls.app_config['font']['max_size'] = value

    @classmethod
    def get_font_max_size(cls) -> int:
        return cls.app_config['font']['max_size']

    """
    Class system variables
    """

    @classmethod
    def set_font_size(cls, value) -> None:
        cls._font_size = value

    @classmethod
    def get_font_size(cls) -> int:
        return cls._font_size

    @classmethod
    def set_prev_font_size(cls, value) -> None:
        cls._prev_font_size = value

    @classmethod
    def get_prev_font_size(cls) -> int:
        return cls._prev_font_size

    @classmethod
    def set_test_mode(cls, value) -> None:
        cls._test_mode = value

    @classmethod
    def get_test_mode(cls) -> bool:
        return cls._test_mode
